import pytest

PUSH_FAILURE_RESULT = {
    "totalSize": 10,
    "done": True,
    "records": [
        {
            "attributes": {
                "type": "PackagePushJob",
                "url": "/services/data/v43.0/sobjects/PackagePushJob/0DX1K000000LcuBWAS",
            },
            "Id": "0DX1K000000LcuBWAS",
            "SubscriberOrganizationKey": "00DFeyLoahEKZqo",
            "PackagePushErrors": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "PackagePushError",
                            "url": "/services/data/v43.0/sobjects/PackagePushError/0DY1K00000013sDWAQ",
                        },
                        "ErrorDetails": None,
                        "ErrorMessage": "This package requires that a newer version of a dependent package be installed before the upgrade can proceed: EDA, Version EDA 1.93.\n\nAsk the subscriber to install the dependent package and then retry the push upgrade.",
                        "ErrorSeverity": "Error",
                        "ErrorTitle": "Dependent Package Conflict",
                        "ErrorType": "IneligibleUpgrade",
                    }
                ],
            },
        },
        {
            "attributes": {
                "type": "PackagePushJob",
                "url": "/services/data/v43.0/sobjects/PackagePushJob/0DX1K000000LcyWWAS",
            },
            "Id": "0DX1K000000LcyWWAS",
            "SubscriberOrganizationKey": "00DhmXPrqbrHoVO",
            "PackagePushErrors": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "PackagePushError",
                            "url": "/services/data/v43.0/sobjects/PackagePushError/0DY1K00000013sIWAQ",
                        },
                        "ErrorDetails": None,
                        "ErrorMessage": "This package requires that a newer version of a dependent package be installed before the upgrade can proceed: EDA, Version EDA 1.93.\n\nAsk the subscriber to install the dependent package and then retry the push upgrade.",
                        "ErrorSeverity": "Error",
                        "ErrorTitle": "Dependent Package Conflict",
                        "ErrorType": "IneligibleUpgrade",
                    }
                ],
            },
        },
        {
            "attributes": {
                "type": "PackagePushJob",
                "url": "/services/data/v43.0/sobjects/PackagePushJob/0DX1K000000LcvOWAS",
            },
            "Id": "0DX1K000000LcvOWAS",
            "SubscriberOrganizationKey": "00DSZGTJkBGDEsA",
            "PackagePushErrors": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "PackagePushError",
                            "url": "/services/data/v43.0/sobjects/PackagePushError/0DY1K00000013s3WAA",
                        },
                        "ErrorDetails": None,
                        "ErrorMessage": "This package requires that a newer version of a dependent package be installed before the upgrade can proceed: EDA, Version EDA 1.93.\n\nAsk the subscriber to install the dependent package and then retry the push upgrade.",
                        "ErrorSeverity": "Error",
                        "ErrorTitle": "Dependent Package Conflict",
                        "ErrorType": "IneligibleUpgrade",
                    }
                ],
            },
        },
        {
            "attributes": {
                "type": "PackagePushJob",
                "url": "/services/data/v43.0/sobjects/PackagePushJob/0DX1K000000Lcv5WAC",
            },
            "Id": "0DX1K000000Lcv5WAC",
            "SubscriberOrganizationKey": "00DhEmESVoRAtGA",
            "PackagePushErrors": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "PackagePushError",
                            "url": "/services/data/v43.0/sobjects/PackagePushError/0DY1K00000013rtWAA",
                        },
                        "ErrorDetails": None,
                        "ErrorMessage": "This package requires that a newer version of a dependent package be installed before the upgrade can proceed: EDA, Version EDA 1.93.\n\nAsk the subscriber to install the dependent package and then retry the push upgrade.",
                        "ErrorSeverity": "Error",
                        "ErrorTitle": "Dependent Package Conflict",
                        "ErrorType": "IneligibleUpgrade",
                    }
                ],
            },
        },
        {
            "attributes": {
                "type": "PackagePushJob",
                "url": "/services/data/v43.0/sobjects/PackagePushJob/0DX1K000000LcxIWAS",
            },
            "Id": "0DX1K000000LcxIWAS",
            "SubscriberOrganizationKey": "00DlwbtJtKxRxQJ",
            "PackagePushErrors": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "PackagePushError",
                            "url": "/services/data/v43.0/sobjects/PackagePushError/0DY1K00000013srWAA",
                        },
                        "ErrorDetails": None,
                        "ErrorMessage": "This package requires that a newer version of a dependent package be installed before the upgrade can proceed: EDA, Version EDA 1.93.\n\nAsk the subscriber to install the dependent package and then retry the push upgrade.",
                        "ErrorSeverity": "Error",
                        "ErrorTitle": "Dependent Package Conflict",
                        "ErrorType": "IneligibleUpgrade",
                    }
                ],
            },
        },
        {
            "attributes": {
                "type": "PackagePushJob",
                "url": "/services/data/v43.0/sobjects/PackagePushJob/0DX1K000000LcxeWAC",
            },
            "Id": "0DX1K000000LcxeWAC",
            "SubscriberOrganizationKey": "00DJDITaMACaQZQ",
            "PackagePushErrors": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "PackagePushError",
                            "url": "/services/data/v43.0/sobjects/PackagePushError/0DY1K00000013s8WAA",
                        },
                        "ErrorDetails": None,
                        "ErrorMessage": "This package requires that a newer version of a dependent package be installed before the upgrade can proceed: EDA, Version EDA 1.93.\n\nAsk the subscriber to install the dependent package and then retry the push upgrade.",
                        "ErrorSeverity": "Error",
                        "ErrorTitle": "Dependent Package Conflict",
                        "ErrorType": "IneligibleUpgrade",
                    }
                ],
            },
        },
        {
            "attributes": {
                "type": "PackagePushJob",
                "url": "/services/data/v43.0/sobjects/PackagePushJob/0DX1K000000LcvhWAC",
            },
            "Id": "0DX1K000000LcvhWAC",
            "SubscriberOrganizationKey": "00DUWEcFgynfLrd",
            "PackagePushErrors": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "PackagePushError",
                            "url": "/services/data/v43.0/sobjects/PackagePushError/0DY1K00000013sNWAQ",
                        },
                        "ErrorDetails": None,
                        "ErrorMessage": "This package requires that a newer version of a dependent package be installed before the upgrade can proceed: EDA, Version EDA 1.93.\n\nAsk the subscriber to install the dependent package and then retry the push upgrade.",
                        "ErrorSeverity": "Error",
                        "ErrorTitle": "Dependent Package Conflict",
                        "ErrorType": "IneligibleUpgrade",
                    }
                ],
            },
        },
        {
            "attributes": {
                "type": "PackagePushJob",
                "url": "/services/data/v43.0/sobjects/PackagePushJob/0DX1K000000Lcy5WAC",
            },
            "Id": "0DX1K000000Lcy5WAC",
            "SubscriberOrganizationKey": "00DbbPwpmXbCdtC",
            "PackagePushErrors": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "PackagePushError",
                            "url": "/services/data/v43.0/sobjects/PackagePushError/0DY1K00000013roWAA",
                        },
                        "ErrorDetails": None,
                        "ErrorMessage": "An unexpected failure was experienced during the upgrade. The subscriber's organization was unaffected. Contact salesforce.com Support through your normal channels and provide the following error number: 731858695-80805 (-151184096).",
                        "ErrorSeverity": "Error",
                        "ErrorTitle": "Unexpected Failure",
                        "ErrorType": "UnclassifiedError",
                    }
                ],
            },
        },
        {
            "attributes": {
                "type": "PackagePushJob",
                "url": "/services/data/v43.0/sobjects/PackagePushJob/0DX1K000000LctfWAC",
            },
            "Id": "0DX1K000000LctfWAC",
            "SubscriberOrganizationKey": "00DTGFxGeEIhFyj",
            "PackagePushErrors": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "PackagePushError",
                            "url": "/services/data/v43.0/sobjects/PackagePushError/0DY1K00000013ryWAA",
                        },
                        "ErrorDetails": None,
                        "ErrorMessage": "This package requires that a newer version of a dependent package be installed before the upgrade can proceed: EDA, Version EDA 1.93.\n\nAsk the subscriber to install the dependent package and then retry the push upgrade.",
                        "ErrorSeverity": "Error",
                        "ErrorTitle": "Dependent Package Conflict",
                        "ErrorType": "IneligibleUpgrade",
                    }
                ],
            },
        },
        {
            "attributes": {
                "type": "PackagePushJob",
                "url": "/services/data/v43.0/sobjects/PackagePushJob/0DX1K000000LcvMWAS",
            },
            "Id": "0DX1K000000LcvMWAS",
            "SubscriberOrganizationKey": "00DmMBUTyLNccMM",
            "PackagePushErrors": {
                "totalSize": 1,
                "done": True,
                "records": [
                    {
                        "attributes": {
                            "type": "PackagePushError",
                            "url": "/services/data/v43.0/sobjects/PackagePushError/0DY1K00000013swWAA",
                        },
                        "ErrorDetails": None,
                        "ErrorMessage": "System.DmlException: Upsert failed. First exception on row 0 with id a4Ig00000001UZhEAM; first error: CANNOT_INSERT_UPDATE_ACTIVATE_ENTITY, hed.TDTM_TriggerHandler: execution of BeforeUpdate\n\ncaused by: System.UnexpectedException: common.exception.SfdcSqlException: ORA-04021: timeout occurred while waiting to lock object \n\n\nClass.gem.TDTM_Glue.getDefaultTdtmConfigTokens: line 66, column 1\nClass.gem.API_HEDA.getDefaultTdtmConfigTokens: line 50, column 1\nClass.hed.TDTM_Config.getTdtmConfig: line 67, column 1\nClass.hed.TDTM_Config.getClassesToCallForObject: line 134, column 1\nClass.hed.TDTM_TriggerHandler.run: line 68, column 1\nClass.hed.TDTM_Global_API.run: line 61, column 1\nTrigger.hed.TDTM_TriggerHandler: line 33, column 1: []\n\nClass.hed.TDTM_Manager.updateDefaultTdtmConfig: line 128, column 1\nClass.hed.TDTM_Global_API.setTdtmConfig: line 102, column 1\nClass.sfal.InstallScript.onInstall: line 5, column 1",
                        "ErrorSeverity": "Error",
                        "ErrorTitle": "Unexpected Error",
                        "ErrorType": "UnclassifiedError",
                    }
                ],
            },
        },
    ],
}


@pytest.fixture()
def push_failure_results():
    return PUSH_FAILURE_RESULT



import io
import zipfile
from unittest import mock

import pytest
import responses

from cumulusci.salesforce_api.exceptions import MetadataApiError
from cumulusci.tasks.salesforce import UninstallPackaged
from cumulusci.tests.util import create_project_config

from .util import create_task


class TestUninstallPackaged:
    @mock.patch("cumulusci.salesforce_api.metadata.ApiRetrievePackaged.__call__")
    def test_get_destructive_changes(self, ApiRetrievePackaged):
        project_config = create_project_config()
        project_config.config["project"]["package"]["name"] = "TestPackage"
        project_config.config["project"]["package"]["api_version"] = "43.0"
        task = create_task(UninstallPackaged, {}, project_config)
        zf = zipfile.ZipFile(io.BytesIO(), "w")
        zf.writestr(
            "TestPackage/objects/Test__c.object",
            '<?xml version="1.0" encoding="UTF-8"?><root />',
        )
        ApiRetrievePackaged.return_value = zf
        result = task._get_destructive_changes()
        assert (
            """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>Test__c</members>
        <name>CustomObject</name>
    </types>
    <version>43.0</version>
</Package>"""
            == result
        )

    @responses.activate
    def test_error_handling(self):
        project_config = create_project_config()
        project_config.config["project"]["package"]["name"] = "TestPackage"
        project_config.config["project"]["package"]["api_version"] = "43.0"
        task = create_task(UninstallPackaged, {}, project_config)

        responses.add(
            method=responses.POST,
            url="https://test.salesforce.com/services/Soap/m/43.0/ORG_ID",
            body="""<?xml version="1.0" encoding="UTF-8"?>
        <soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"
        xmlns="http://soap.sforce.com/2006/04/metadata" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
        <soapenv:Body><checkRetrieveStatusResponse><result><done>true</done>
        <errorMessage>INVALID_CROSS_REFERENCE_KEY: No package named &apos;Abacus&apos; found</errorMessage>
        <errorStatusCode>UNKNOWN_EXCEPTION</errorStatusCode><id>09S6g000007KGwVEAW</id><status>Failed</status>
        <success>false</success><zipFile xsi:nil="true"/></result>
        </checkRetrieveStatusResponse></soapenv:Body></soapenv:Envelope>""",
            status=200,
            content_type="text/xml; charset=utf-8",
        )

        with pytest.raises(MetadataApiError) as e:
            task()
        assert "No package" in str(e.value)



import json
import os
import re
from configparser import ConfigParser
from contextlib import contextmanager
from distutils.version import LooseVersion
from io import StringIO
from itertools import chain
from pathlib import Path

API_VERSION_RE = re.compile(r"^\d\d+\.0$")

import github3

from cumulusci.core.config import BaseTaskFlowConfig
from cumulusci.core.exceptions import (
    ConfigError,
    GithubException,
    KeychainNotFound,
    NamespaceNotFoundError,
    NotInProject,
    ProjectConfigNotFound,
)
from cumulusci.core.github import (
    catch_common_github_auth_errors,
    find_previous_release,
    get_github_api_for_repo,
)
from cumulusci.core.source import GitHubSource, LocalFolderSource, NullSource
from cumulusci.core.utils import merge_config
from cumulusci.utils.fileutils import open_fs_resource
from cumulusci.utils.git import current_branch, git_path, split_repo_url
from cumulusci.utils.yaml.cumulusci_yml import cci_safe_load


class BaseProjectConfig(BaseTaskFlowConfig):
    """Base class for a project's configuration which extends the global config"""

    config_filename = "cumulusci.yml"

    def __init__(self, universal_config_obj, config=None, *args, **kwargs):
        self.universal_config_obj = universal_config_obj
        self.keychain = None

        # optionally pass in a repo_info dict
        self._repo_info = kwargs.pop("repo_info", None)

        if not config:
            config = {}

        # Initialize the dictionaries for the individual configs
        self.config_project = {}
        self.config_project_local = {}
        self.config_additional_yaml = {}

        # optionally pass in a kwarg named 'additional_yaml' that will
        # be added to the YAML merge stack.
        self.additional_yaml = None
        if "additional_yaml" in kwargs:
            self.additional_yaml = kwargs.pop("additional_yaml")

        # initialize map of project configs referenced from an external source
        self.source = NullSource()
        self.included_sources = kwargs.pop("included_sources", {})

        super(BaseProjectConfig, self).__init__(config=config)

    @property
    def config_project_local_path(self):
        path = Path(self.project_local_dir) / self.config_filename
        if path.is_file():
            return str(path)

    def _load_config(self):
        """Loads the configuration from YAML, if no override config was passed in initially."""

        if (
            self.config
        ):  # any config being pre-set at init will short circuit out, but not a plain {}
            return

        # Verify that we're in a project
        repo_root = self.repo_root
        if not repo_root:
            raise NotInProject(
                "No git repository was found in the current path. You must be in a git repository to set up and use CCI for a project."
            )

        # Verify that the project's root has a config file
        if not self.config_project_path:
            raise ProjectConfigNotFound(
                f"The file {self.config_filename} was not found in the repo root: {repo_root}. Are you in a CumulusCI Project directory?"
            )

        # Load the project's yaml config file
        project_config = cci_safe_load(self.config_project_path, logger=self.logger)

        if project_config:
            self.config_project.update(project_config)

        # Load the local project yaml config file if it exists
        if self.config_project_local_path:
            local_config = cci_safe_load(
                self.config_project_local_path, logger=self.logger
            )
            if local_config:
                self.config_project_local.update(local_config)

        # merge in any additional yaml that was passed along
        if self.additional_yaml:
            additional_yaml_config = cci_safe_load(
                StringIO(self.additional_yaml),
                self.config_project_path,
                logger=self.logger,
            )
            if additional_yaml_config:
                self.config_additional_yaml.update(additional_yaml_config)

        self.config = merge_config(
            {
                "universal_config": self.config_universal,
                "global_config": self.config_global,
                "project_config": self.config_project,
                "project_local_config": self.config_project_local,
                "additional_yaml": self.config_additional_yaml,
            }
        )

        self._validate_config()

    def _validate_config(self):
        """Performs validation checks on the configuration"""
        self._validate_package_api_format()

    def _validate_package_api_format(self):
        api_version = str(self.project__package__api_version)

        if not API_VERSION_RE.match(api_version):
            message = (
                f"Package API Version must be in the form 'XX.0', found: {api_version}"
            )
            raise ConfigError(message)

    @property
    def config_global(self):
        return self.universal_config_obj.config_global

    @property
    def config_universal(self):
        return self.universal_config_obj.config_universal

    @property
    def repo_info(self):
        if self._repo_info is not None:
            return self._repo_info

        # Detect if we are running in a CI environment and get repo info
        # from env vars for the enviornment instead of .git files
        info = {"ci": None}

        # Make sure that the CUMULUSCI_AUTO_DETECT environment variable is
        # set before trying to auto-detect anything from the environment
        if not os.environ.get("CUMULUSCI_AUTO_DETECT"):
            self._repo_info = info
            return self._repo_info

        # Heroku CI
        heroku_ci = os.environ.get("HEROKU_TEST_RUN_ID")
        if heroku_ci:
            info = {
                "branch": os.environ.get("HEROKU_TEST_RUN_BRANCH"),
                "commit": os.environ.get("HEROKU_TEST_RUN_COMMIT_VERSION"),
                "ci": "heroku",
                "root": "/app",
            }

        # Other CI environment implementations can be implemented here...

        self._apply_repo_env_var_overrides(info)

        if info["ci"]:
            self._validate_required_git_info(info)

        if len(info) > 1:
            self._log_detected_overrides_as_warning(info)

        self._repo_info = info
        return self._repo_info

    def _apply_repo_env_var_overrides(self, info):
        """Apply CUMULUSCI_REPO_* environment variables last so they can
        override and fill in missing values from the CI environment"""
        self._override_repo_env_var("CUMULUSCI_REPO_BRANCH", "branch", info)
        self._override_repo_env_var("CUMULUSCI_REPO_COMMIT", "commit", info)
        self._override_repo_env_var("CUMULUSCI_REPO_ROOT", "root", info)

        repo_url = os.environ.get("CUMULUSCI_REPO_URL")
        if repo_url:
            if repo_url != info.get("url"):
                self.logger.info(
                    "CUMULUSCI_REPO_URL found, using its value as the repo url, owner, and name"
                )
            url_info = {}
            url_info["owner"], url_info["name"] = split_repo_url(repo_url)
            url_info["url"] = repo_url
            info.update(url_info)

    def _override_repo_env_var(self, repo_env_var, local_var, info):
        repo_env_var = os.environ.get(repo_env_var)
        if repo_env_var:
            if repo_env_var != info.get(local_var):
                self.logger.info("{} found, using its value for configuration.")
            info[local_var] = repo_env_var

    def _validate_required_git_info(self, info):
        """Ensures that we have the required git info or throw a ConfigError"""
        validate = {
            # <key>: <env var to manually override>
            "branch": "CUMULUSCI_REPO_BRANCH",
            "commit": "CUMULUSCI_REPO_COMMIT",
            "name": "CUMULUSCI_REPO_URL",
            "owner": "CUMULUSCI_REPO_URL",
            "root": "CUMULUSCI_REPO_ROOT",
            "url": "CUMULUSCI_REPO_URL",
        }
        for key, env_var in list(validate.items()):
            if key not in info or not info[key]:
                message = f"Detected CI on {info['ci']} but could not determine the repo {key}"
                if env_var:
                    message += f". You can manually pass in the {key} "
                    message += f" with the {env_var} environment variable."
                raise ConfigError(message)

    def _log_detected_overrides_as_warning(self, info):
        self.logger.info("")
        self.logger.warning("Using environment variables to override repo info:")
        keys = list(info.keys())
        keys.sort()
        for key in keys:
            self.logger.warning(f"  {key}: {info[key]}")
        self.logger.info("")

    def git_config_remote_origin_url(self):
        """Returns the url under the [remote origin]
        section of the .git/config file. Returns None
        if .git/config file not present or no matching
        line is found."""
        config = ConfigParser(strict=False)
        try:
            config.read(git_path(self.repo_root, "config"))
            url = config['remote "origin"']["url"]
        except (KeyError, TypeError):
            url = None
        return url

    @property
    def repo_root(self):
        path = self.repo_info.get("root")
        if path:
            return path

        path = Path.cwd().resolve()
        paths = chain((path,), path.parents)
        for path in paths:
            if (path / ".git").is_dir():
                return str(path)

    @property
    def repo_name(self):
        name = self.repo_info.get("name")
        if name:
            return name

        if not self.repo_root:
            return

        url_line = self.git_config_remote_origin_url()
        return split_repo_url(url_line)[1]

    @property
    def repo_url(self):
        url = self.repo_info.get("url")
        if url:
            return url

        if not self.repo_root:
            return

        url = self.git_config_remote_origin_url()
        return url

    @property
    def repo_owner(self):
        owner = self.repo_info.get("owner")
        if owner:
            return owner

        if not self.repo_root:
            return

        url_line = self.git_config_remote_origin_url()
        return split_repo_url(url_line)[0]

    @property
    def repo_branch(self):
        branch = self.repo_info.get("branch")
        if branch:
            return branch

        if not self.repo_root:
            return

        return current_branch(self.repo_root)

    @property
    def repo_commit(self):
        commit = self.repo_info.get("commit")
        if commit:
            return commit

        if not self.repo_root:
            return

        branch = self.repo_branch
        if not branch:
            return

        join_args = [self.repo_root, ".git", "refs", "heads"]
        join_args.extend(branch.split("/"))
        commit_file = os.path.join(*join_args)

        commit_sha = None
        if os.path.isfile(commit_file):
            with open(commit_file, "r") as f:
                commit_sha = f.read().strip()
        else:
            packed_refs_path = os.path.join(self.repo_root, ".git", "packed-refs")
            with open(packed_refs_path, "r") as f:
                for line in f:
                    parts = line.split(" ")
                    if len(parts) == 1:
                        # Skip lines showing the commit sha of a tag on the
                        # preceeding line
                        continue
                    if parts[1].replace("refs/remotes/origin/", "").strip() == branch:
                        commit_sha = parts[0]
                        break

        return commit_sha

    def get_github_api(self, owner=None, repo=None):
        return get_github_api_for_repo(
            self.keychain, owner or self.repo_owner, repo or self.repo_name
        )

    def get_repo(self):
        repo = self.get_github_api(self.repo_owner, self.repo_name).repository(
            self.repo_owner, self.repo_name
        )
        if repo is None:
            raise GithubException(
                f"Github repository not found or not authorized. ({self.repo_url})"
            )
        return repo

    # TODO: These methods are duplicative with `find_latest_release()`
    def get_latest_tag(self, beta=False):
        """Query Github Releases to find the latest production or beta tag"""
        repo = self.get_repo()
        if not beta:
            try:
                release = repo.latest_release()
            except github3.exceptions.NotFoundError:
                raise GithubException(f"No release found for repo {self.repo_url}")
            prefix = self.project__git__prefix_release
            if not release.tag_name.startswith(prefix):
                return self._get_latest_tag_for_prefix(repo, prefix)
            return release.tag_name
        else:
            return self._get_latest_tag_for_prefix(repo, self.project__git__prefix_beta)

    def _get_latest_tag_for_prefix(self, repo, prefix):
        for release in repo.releases():
            if not release.tag_name.startswith(prefix):
                continue
            return release.tag_name
        raise GithubException(
            f"No release found for {self.repo_url} with tag prefix {prefix}"
        )

    def get_latest_version(self, beta=False):
        """Query Github Releases to find the latest production or beta release"""
        tag = self.get_latest_tag(beta)
        version = self.get_version_for_tag(tag)
        if version is not None:
            return LooseVersion(version)

    def get_previous_version(self):
        """Query GitHub releases to find the previous production release"""
        repo = self.get_repo()
        release = find_previous_release(repo, self.project__git__prefix_release)
        if release is not None:
            return LooseVersion(self.get_version_for_tag(release.tag_name))

    @property
    def config_project_path(self):
        if not self.repo_root:
            return
        path = Path(self.repo_root) / self.config_filename
        if path.is_file():
            return str(path)

    @property
    def project_local_dir(self):
        """location of the user local directory for the project
        e.g., ~/.cumulusci/NPSP-Extension-Test/"""

        # depending on where we are in bootstrapping the UniversalConfig
        # the canonical projectname could be located in one of two places
        if self.project__name:
            name = self.project__name
        else:
            name = self.config_project.get("project", {}).get("name", "")

        path = str(self.universal_config_obj.cumulusci_config_dir / name)
        if not os.path.isdir(path):
            os.makedirs(path)
        return path

    @property
    def default_package_path(self):
        if self.project__source_format == "sfdx":
            relpath = "force-app"
            for pkg in self.sfdx_project_config.get("packageDirectories", []):
                if pkg.get("default"):
                    relpath = pkg["path"]
        else:
            relpath = "src"
        return Path(self.repo_root, relpath).resolve()

    @property
    def sfdx_project_config(self):
        with open(
            Path(self.repo_root) / "sfdx-project.json", "r", encoding="utf-8"
        ) as f:
            config = json.load(f)
        return config

    def get_tag_for_version(self, version, prefix=None):
        """Given a version, returns the appropriate tag name to use.
        If a prefix is not specified, we infer beta vs. release based
        on 1GP beta version naming conventions."""
        if prefix:
            tag_name = prefix + version
        elif "(Beta" in version:
            tag_version = version.replace(" (", "-").replace(")", "").replace(" ", "_")
            tag_name = self.project__git__prefix_beta + tag_version
        else:
            tag_name = self.project__git__prefix_release + version
        return tag_name

    def get_version_for_tag(self, tag, prefix_beta=None, prefix_release=None):
        if prefix_beta is None:
            prefix_beta = self.project__git__prefix_beta
        if prefix_release is None:
            prefix_release = self.project__git__prefix_release
        if tag.startswith(prefix_beta):
            version = tag.replace(prefix_beta, "")
            if "-Beta_" in version:
                # Beta tags are expected to be like "beta/1.0-Beta_1"
                # which is returned as "1.0 (Beta 1)"
                return version.replace("-", " (").replace("_", " ") + ")"
            else:
                return
        elif tag.startswith(prefix_release):
            return tag.replace(prefix_release, "")

    def set_keychain(self, keychain):
        self.keychain = keychain

    def _check_keychain(self):
        if not self.keychain:
            raise KeychainNotFound(
                "Could not find config.keychain. You must call "
                + "config.set_keychain(keychain) before accessing orgs"
            )

    @catch_common_github_auth_errors
    def get_repo_from_url(self, url):
        owner, name = split_repo_url(url)
        return self.get_github_api(owner, name).repository(owner, name)

    def get_task(self, name):
        """Get a TaskConfig by task name

        If the name has a colon, look for it in a different project config.
        """
        if ":" in name:
            ns, name = name.split(":")
            other_config = self.get_namespace(ns)
            task_config = other_config.get_task(name)
            task_config.project_config = other_config
        else:
            task_config = super().get_task(name)
            task_config.project_config = self
        return task_config

    def get_flow(self, name):
        """Get a FlowConfig by flow name

        If the name has a colon, look for it in a different project config.
        """
        if ":" in name:
            ns, name = name.split(":")
            other_config = self.get_namespace(ns)
            flow_config = other_config.get_flow(name)
            flow_config.name = name
            flow_config.project_config = other_config
        else:
            flow_config = super().get_flow(name)
            flow_config.name = name
            flow_config.project_config = self
        return flow_config

    def get_namespace(self, ns):
        """Look up another project config by its name in the `sources` config.

        Also makes sure the project has been fetched, if it's from an external source.
        """
        spec = getattr(self, f"sources__{ns}")
        if spec is None:
            raise NamespaceNotFoundError(f"Namespace not found: {ns}")
        return self.include_source(spec)

    def include_source(self, spec):
        """Make sure a project has been fetched from its source.

        `spec` is a dict which contains different keys depending on the type of source:

        - `github` indicates a GitHubSource
        - `path` indicates a LocalFolderSource

        This either fetches the project code and constructs its project config,
        or returns a project config that was previously loaded with the same spec.
        """
        frozenspec = tuple(spec.items())
        if frozenspec in self.included_sources:
            project_config = self.included_sources[frozenspec]
        else:
            if "github" in spec:
                source = GitHubSource(self, spec)
            elif "path" in spec:
                source = LocalFolderSource(self, spec)
            else:
                raise Exception(f"Not sure how to load project: {spec}")
            self.logger.info(f"Fetching from {source}")
            project_config = source.fetch()
            project_config.set_keychain(self.keychain)
            project_config.source = source
            self.included_sources[frozenspec] = project_config
        return project_config

    def construct_subproject_config(self, **kwargs):
        """Construct another project config for an external source"""
        return self.__class__(
            self.universal_config_obj, included_sources=self.included_sources, **kwargs
        )

    def relpath(self, path):
        """Convert path to be relative to the project repo root."""
        return os.path.relpath(os.path.join(self.repo_root, path))

    @property
    def cache_dir(self):
        "A project cache which is on the local filesystem. Prefer open_cache where possible."
        assert self.repo_root
        cache_dir = Path(self.repo_root, ".cci")
        cache_dir.mkdir(exist_ok=True)
        return cache_dir

    @contextmanager
    def open_cache(self, cache_name):
        "A context managed PyFilesystem-based cache which could theoretically be on any filesystem."
        with open_fs_resource(self.cache_dir) as cache_dir:
            cache = cache_dir / cache_name
            cache.mkdir(exist_ok=True)
            yield cache



import io
import logging
import zipfile
from unittest import mock

import pydantic
import pytest

from cumulusci.core.dependencies.dependencies import (
    GitHubDynamicDependency,
    PackageNamespaceVersionDependency,
    PackageVersionIdDependency,
    UnmanagedGitHubRefDependency,
)
from cumulusci.core.dependencies.resolvers import (
    DependencyResolutionStrategy,
    get_resolver_stack,
)
from cumulusci.core.exceptions import DependencyParseError, TaskOptionsError
from cumulusci.core.flowrunner import StepSpec
from cumulusci.tasks.salesforce.update_dependencies import UpdateDependencies
from cumulusci.tests.util import create_project_config

from .util import create_task


def make_fake_zipfile(*args, **kw):
    return zipfile.ZipFile(io.BytesIO(), "w")


def test_init_options_base():
    project_config = create_project_config()

    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [
                {
                    "namespace": "ns",
                    "version": "1.0",
                },
                {"version_id": "04t000000000000"},
                {"github": "https://github.com/Test/TestRepo"},
            ],
            "resolution_strategy": "production",
        },
        project_config=project_config,
    )

    assert task.dependencies == [
        PackageNamespaceVersionDependency(namespace="ns", version="1.0"),
        PackageVersionIdDependency(version_id="04t000000000000"),
        GitHubDynamicDependency(github="https://github.com/Test/TestRepo"),
    ]
    assert task.resolution_strategy == get_resolver_stack(project_config, "production")


def test_init_options_error_bad_dependencies():
    with pytest.raises(DependencyParseError):
        create_task(
            UpdateDependencies,
            {
                "dependencies": [
                    {
                        "namespace": "ns",
                        "version_id": "04t000000000000",
                    }
                ]
            },
        )


def test_init_options_warns_deprecated_options(caplog):
    with caplog.at_level(logging.INFO):
        create_task(
            UpdateDependencies,
            {
                "dependencies": [
                    {
                        "namespace": "ns",
                        "version": "1.0",
                    }
                ],
                "allow_uninstalls": False,
                "include_beta": True,
            },
        )

        assert "no longer supported" in caplog.text
        assert "Use resolution strategies instead" in caplog.text


def test_init_options_error_bad_ignore_dependencies():
    with pytest.raises(TaskOptionsError):
        create_task(
            UpdateDependencies,
            {
                "dependencies": [
                    {
                        "namespace": "ns",
                        "version": "1.0",
                    }
                ],
                "ignore_dependencies": [{"foo": "bar"}],
            },
        )


def test_init_options_uses_include_beta_strategy_for_include_beta_true():
    org_config = mock.Mock()
    org_config.scratch = True

    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [
                {
                    "namespace": "ns",
                    "version": "1.0",
                }
            ],
            "include_beta": True,
            "resolution_strategy": "production",
        },
        org_config=org_config,
    )

    assert DependencyResolutionStrategy.BETA_RELEASE_TAG in task.resolution_strategy


def test_init_options_removes_beta_resolver_for_include_beta_false():
    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [
                {
                    "namespace": "ns",
                    "version": "1.0",
                }
            ],
            "resolution_strategy": "include_beta",
            "include_beta": False,
        },
    )

    assert DependencyResolutionStrategy.BETA_RELEASE_TAG not in task.resolution_strategy


def test_init_options_removes_2gp_resolver_for_prefer_2gp_false():
    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [
                {
                    "namespace": "ns",
                    "version": "1.0",
                }
            ],
            "resolution_strategy": "include_beta",
            "prefer_2gp_from_release_branch": False,
        },
    )

    assert (
        DependencyResolutionStrategy.COMMIT_STATUS_RELEASE_BRANCH
        not in task.resolution_strategy
    )


def test_init_options_removes_unsafe_resolvers_persistent_org():
    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [
                {
                    "namespace": "ns",
                    "version": "1.0",
                }
            ],
            "resolution_strategy": "include_beta",
        },
    )
    task.org_config = mock.Mock()
    task.org_config.scratch = False

    assert DependencyResolutionStrategy.BETA_RELEASE_TAG not in task.resolution_strategy
    assert (
        DependencyResolutionStrategy.COMMIT_STATUS_RELEASE_BRANCH
        not in task.resolution_strategy
    )


def test_run_task_gets_static_dependencies_and_installs():
    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [
                {
                    "namespace": "ns",
                    "version": "1.0",
                },
                {"version_id": "04t000000000000"},
            ],
            "resolution_strategy": "production",
            "ignore_dependencies": [{"github": "https://github.com/Test/TestRepo"}],
            "security_type": "PUSH",
        },
    )

    task._install_dependency = mock.Mock()
    task()

    task._install_dependency.assert_has_calls(
        [
            mock.call(PackageNamespaceVersionDependency(namespace="ns", version="1.0")),
            mock.call(PackageVersionIdDependency(version_id="04t000000000000")),
        ]
    )


def test_run_task_gets_static_dependencies_and_installs__packages_only():
    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [
                {
                    "namespace": "ns",
                    "version": "1.0",
                },
                {
                    "github": "https://github.com/TestRepo/Test",
                    "ref": "aaaa",
                    "subfolder": "foo",
                },
            ],
            "resolution_strategy": "production",
            "security_type": "PUSH",
            "packages_only": True,
        },
    )

    task._install_dependency = mock.Mock()
    task()

    assert task._install_dependency.call_count == 1
    task._install_dependency.assert_has_calls(
        [
            mock.call(PackageNamespaceVersionDependency(namespace="ns", version="1.0")),
        ]
    )


def test_run_task_exits_no_dependencies():
    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [],
            "resolution_strategy": "production",
            "ignore_dependencies": [{"github": "https://github.com/Test/TestRepo"}],
            "security_type": "PUSH",
        },
    )

    task._install_dependency = mock.Mock()
    task()

    task._install_dependency.assert_not_called()


@mock.patch(
    "cumulusci.core.dependencies.dependencies.install_package_by_namespace_version"
)
def test_install_dependency_installs_managed_package(
    install_package_by_namespace_version,
):
    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [
                {
                    "namespace": "ns",
                    "version": "1.0",
                }
            ]
        },
    )
    task.org_config = mock.Mock()
    task.org_config.installed_packages = {}
    task.org_config.has_minimum_package_version.return_value = False

    task._install_dependency(task.dependencies[0])
    install_package_by_namespace_version.assert_called_once_with(
        task.project_config,
        task.org_config,
        "ns",
        "1.0",
        mock.ANY,
        retry_options=mock.ANY,  # Ignore the options
    )


def test_install_dependency_installs_unmanaged():
    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [
                {
                    "zip_url": "http://example.com/foo",
                }
            ]
        },
    )
    task.dependencies[0].__config__.extra = pydantic.Extra.allow
    task.dependencies[0].install = mock.Mock()
    task.org_config = mock.Mock()

    task._install_dependency(task.dependencies[0])
    task.dependencies[0].install.assert_called_once_with(
        task.project_config, task.org_config
    )


def test_run_task__bad_security_type():
    with pytest.raises(TaskOptionsError):
        create_task(
            UpdateDependencies,
            {
                "security_type": "BOGUS",
                "dependencies": [
                    {
                        "namespace": "ns",
                        "version": "1.0",
                    },
                ],
            },
        )


@mock.patch("cumulusci.tasks.salesforce.update_dependencies.get_static_dependencies")
def test_freeze(get_static_dependencies):
    get_static_dependencies.return_value = [
        PackageNamespaceVersionDependency(namespace="ns", version="1.0"),
        UnmanagedGitHubRefDependency(
            github="https://github.com/SFDO-Tooling/CumulusCI-Test",
            ref="abcdef",
            subfolder="src",
        ),
    ]
    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [
                {
                    "namespace": "ns",
                    "version": "1.0",
                },
                {
                    "github": "https://github.com/SFDO-Tooling/CumulusCI-Test",
                    "ref": "abcdef",
                    "subfolder": "src",
                },
            ]
        },
    )
    step = StepSpec(1, "test_task", task.task_config, None, task.project_config)
    steps = task.freeze(step)

    assert [
        {
            "is_required": True,
            "kind": "managed",
            "name": "Install ns 1.0",
            "path": "test_task.1",
            "step_num": "1.1",
            "source": None,
            "task_class": None,
            "task_config": {
                "options": {
                    "dependencies": [{"namespace": "ns", "version": "1.0"}],
                    "security_type": "FULL",
                    "packages_only": False,
                },
                "checks": [],
            },
        },
        {
            "is_required": True,
            "kind": "metadata",
            "name": "Deploy https://github.com/SFDO-Tooling/CumulusCI-Test",
            "path": "test_task.2",
            "step_num": "1.2",
            "source": None,
            "task_class": None,
            "task_config": {
                "options": {
                    "dependencies": [
                        {
                            "ref": "abcdef",
                            "github": "https://github.com/SFDO-Tooling/CumulusCI-Test",
                            "subfolder": "src",
                        }
                    ],
                    "security_type": "FULL",
                    "packages_only": False,
                },
                "checks": [],
            },
        },
    ] == steps


@mock.patch("cumulusci.tasks.salesforce.update_dependencies.get_static_dependencies")
def test_freeze__packages_only(get_static_dependencies):
    get_static_dependencies.return_value = [
        PackageNamespaceVersionDependency(namespace="ns", version="1.0"),
        UnmanagedGitHubRefDependency(
            github="https://github.com/SFDO-Tooling/CumulusCI-Test",
            ref="abcdef",
            subfolder="src",
        ),
    ]
    task = create_task(
        UpdateDependencies,
        {
            "dependencies": [
                {
                    "namespace": "ns",
                    "version": "1.0",
                },
                {
                    "github": "https://github.com/SFDO-Tooling/CumulusCI-Test",
                    "ref": "abcdef",
                    "subfolder": "src",
                },
            ],
            "packages_only": True,
        },
    )
    step = StepSpec(1, "test_task", task.task_config, None, task.project_config)
    steps = task.freeze(step)

    assert [
        {
            "is_required": True,
            "kind": "managed",
            "name": "Install ns 1.0",
            "path": "test_task.1",
            "step_num": "1.1",
            "source": None,
            "task_class": None,
            "task_config": {
                "options": {
                    "dependencies": [{"namespace": "ns", "version": "1.0"}],
                    "security_type": "FULL",
                    "packages_only": True,
                },
                "checks": [],
            },
        }
    ] == steps



import logging

import robot.api.logger
from robot.libraries.BuiltIn import BuiltIn

from cumulusci.cli.runtime import CliRuntime
from cumulusci.core.config import TaskConfig
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.tasks import CURRENT_TASK
from cumulusci.core.utils import import_global
from cumulusci.robotframework.utils import set_pdb_trace
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection
from cumulusci.tasks.robotframework.robotframework import Robot


class CumulusCI(object):
    """Library for accessing CumulusCI for the local git project

    This library allows Robot Framework tests to access credentials to a
    Salesforce org created by CumulusCI, including Scratch Orgs.  It also
    exposes the core logic of CumulusCI including interactions with the
    Salesforce API's and project specific configuration including custom
    and customized tasks and flows.

    Initialization requires a single argument, the org name for the target
    CumulusCI org.  If running your tests via cci's robot task (recommended),
    you can initialize the library in your tests taking advantage of the
    variable set by the robot task:
    | ``*** Settings ***``
    |
    | Library  cumulusci.robotframework.CumulusCI  ${ORG}

    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"

    def __init__(self, org_name=None):
        if not org_name:
            org_name = "dev"
        self.org_name = org_name
        self._project_config = None
        self._org = None

        # Turn off info logging of all http requests
        logging.getLogger("requests.packages.urllib3.connectionpool").setLevel(
            logging.WARN
        )

    @property
    def project_config(self):
        if self._project_config is None:
            if getattr(CURRENT_TASK, "stack", None) and isinstance(
                CURRENT_TASK.stack[0], Robot
            ):
                # If CumulusCI is running a task, use that task's config
                return CURRENT_TASK.stack[0].project_config
            else:
                robot.api.logger.console("Initializing CumulusCI config\n")
                self._project_config = CliRuntime().project_config
        return self._project_config

    def set_project_config(self, project_config):
        self._project_config = project_config

    @property
    def keychain(self):
        return self.project_config.keychain

    @property
    def org(self):
        if self._org is None:
            if getattr(CURRENT_TASK, "stack", None) and isinstance(
                CURRENT_TASK.stack[0], Robot
            ):
                # If CumulusCI is running a task, use that task's org
                return CURRENT_TASK.stack[0].org_config
            else:
                self._org = self.keychain.get_org(self.org_name)
        return self._org

    @property
    def sf(self):
        return self._init_api()

    @property
    def tooling(self):
        return self._init_api("tooling/")

    def set_login_url(self):
        """Sets the LOGIN_URL variable in the suite scope which will
        automatically log into the target Salesforce org.

        Typically, this is run during Suite Setup
        """
        BuiltIn().set_suite_variable("${LOGIN_URL}", self.org.start_url)

    def get_org_info(self):
        """Returns a dictionary of the org information for the current target
        Salesforce org
        """
        return self.org.config

    def login_url(self, org=None, **userfields):
        """Returns the login url which will automatically log into the target
        Salesforce org.  By default, the org_name passed to the library
        constructor is used but this can be overridden with the org option
        to log into a different org.

        If userfields are provided, the username and access token
        for the given user will be used. If not provided, the access token
        for the org's default user will be used.

        Example:

        | ${login url}=  Login URL  alias=dadvisor

        """
        org = self.org if org is None else self.keychain.get_org(org)
        if userfields:
            access_token = org.get_access_token(**userfields)
            login_url = f"{org.instance_url}/secur/frontdoor.jsp?sid={access_token}"
            return login_url
        else:
            return org.start_url

    def get_community_info(self, community_name, key=None, force_refresh=False):
        """This keyword uses the Salesforce API to get information about a community.

        This keyword requires the exact community name as its first argumment.

        - If no key is given, all of the information returned by the API will be
          returned by this keyword in the form of a dictionary
        - If a key is given, only the value for that key will be returned.

        Some of the supported keys include name, siteUrl, and
        loginUrl. For a comprehensive list see the
        [https://developer.salesforce.com/docs/atlas.en-us.chatterapi.meta/chatterapi/connect_responses_community.htm|API documentation],
        or call this keyword without the key argument and examine the
        results.

        An API call will be made the first time this keyword is used, and
        the return values will be cached. Subsequent calls will not call
        the API unless the requested community name is not in the cached
        results, or unless the force_refresh parameter is set to True.
        """
        community_info = self.org.get_community_info(
            community_name, force_refresh=force_refresh
        )
        if key is None:
            return community_info
        else:
            if key not in community_info:
                raise Exception("Invalid key '{}'".format(key))
            return community_info[key]

    def get_namespace_prefix(self, package=None):
        """Returns the namespace prefix (including __) for the specified package name.
        (Defaults to project__package__name_managed from the current project config.)

        Returns an empty string if the package is not installed as a managed package.
        """
        result = ""
        if package is None:
            package = self.project_config.project__package__name_managed
        packages = self.tooling.query(
            "SELECT SubscriberPackage.NamespacePrefix, SubscriberPackage.Name "
            "FROM InstalledSubscriberPackage"
        )
        match = [
            p for p in packages["records"] if p["SubscriberPackage"]["Name"] == package
        ]
        if match:
            result = match[0]["SubscriberPackage"]["NamespacePrefix"] + "__"
        return result

    def run_task(self, task_name, **options):
        """Runs a named CumulusCI task for the current project with optional
        support for overriding task options via kwargs.

        Note: task_name can be prefixed with the name of another project,
        just the same as when running the task from the command line. The other
        project needs to have been defined in the 'sources' section of cumulusci.yml.

        The task output will appear in the robot log.

        Examples:
        | =Keyword= | =task_name= | =task_options=             | =comment=                        |
        | Run Task  | deploy      |                            | Run deploy with standard options |
        | Run Task  | deploy      | path=path/to/some/metadata | Run deploy with custom path      |
        | Run task  | npsp:deploy_rd2_config  |                | Run the deploy_rd2_config task from the NPSP project |
        """
        task_config = self.project_config.get_task(task_name)
        class_path = task_config.class_path
        task = self._init_task(class_path, options, task_config)
        return self._run_task(task)

    def run_task_class(self, class_path, **options):
        """Runs a CumulusCI task class with task options via kwargs.

        Use this keyword to run logic from CumulusCI tasks which have not
        been configured in the project's cumulusci.yml file.  This is
        most useful in cases where a test needs to use task logic for
        logic unique to the test and thus not worth making into a named
        task for the project

        The task output will appear in the robot log.

        Examples:
        | =Keyword=      | =task_class=                     | =task_options=                            |
        | Run Task Class | cumulusci.task.utils.DownloadZip | url=http://test.com/test.zip dir=test_zip |
        """
        task = self._init_task(class_path, options, TaskConfig())
        return self._run_task(task)

    def _init_api(self, base_url=None):
        client = get_simple_salesforce_connection(self.project_config, self.org)
        if base_url is not None:
            client.base_url += base_url
        return client

    def _init_task(self, class_path, options, task_config):
        task_class = import_global(class_path)
        task_config = self._parse_task_options(options, task_class, task_config)
        # Python deprecated the logger method "warn" in favor of
        # "warning". Robot didn't get the memo and only has a "warn"
        # method. Some tasks use "warning", so this makes sure the
        # robot logger can handle that.
        if not hasattr(robot.api.logger, "warning"):
            robot.api.logger.warning = robot.api.logger.warn
        task = task_class(
            task_config.project_config or self.project_config,
            task_config,
            org_config=self.org,
            logger=robot.api.logger,
        )
        return task

    def _parse_task_options(self, options, task_class, task_config):
        if "options" not in task_config.config:
            task_config.config["options"] = {}
        # Parse options and add to task config
        if options:
            for name, value in options.items():
                # Validate the option
                if name not in task_class.task_options:
                    raise TaskOptionsError(
                        'Option "{}" is not available for task {}'.format(
                            name, task_class
                        )
                    )

                # Override the option in the task config
                task_config.config["options"][name] = value

        return task_config

    def _run_task(self, task):
        task()
        return task.return_values

    def debug(self):
        """Pauses execution and enters the Python debugger."""
        set_pdb_trace()



import pytest

from cumulusci.utils.http.multi_request import CompositeParallelSalesforce


class TestCompositeParallelSalesforce:
    @pytest.mark.vcr()
    def test_composite_parallel_salesforce(self, sf):
        sf.Account.create(
            {"Name": "Smith Corp."},
        )
        requests = [
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Id FROM Account LIMIT 1",
            },
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
            },
        ] * 2
        with CompositeParallelSalesforce(sf, 5, max_workers=1) as cpsf:
            results = cpsf.do_composite_requests(requests)
        assert len(list(results)) == len(requests)
        for result in results:
            assert result["body"]
            assert result["body"]["done"]
            assert result["body"]["records"][0]["attributes"]

    @pytest.mark.vcr()
    def test_not_opened(self, sf):
        cpsf = CompositeParallelSalesforce(sf, 5, max_workers=1)
        with pytest.raises(AssertionError):
            cpsf.do_composite_requests([])

    @pytest.mark.vcr()
    def test_empty(self, sf):
        with CompositeParallelSalesforce(sf, 5, max_workers=1) as cpsf:
            results = cpsf.do_composite_requests([])
            assert list(results) == []

    # don't re-record this one because you'll need to fiddle
    # with the date
    @pytest.mark.vcr(record_mode="none")
    def test_http_headers(self, sf):
        requests = [
            {
                "method": "GET",
                "url": "/services/data/v50.0/sobjects",
                "httpHeaders": {"If-Modified-Since": "Thu, 03 Sep 2020 21:35:07 GMT"},
            },
        ] * 3
        with CompositeParallelSalesforce(sf, 4, max_workers=1) as cpsf:
            results = cpsf.do_composite_requests(requests)
        assert results[0]["httpStatusCode"] == 304

    @pytest.mark.vcr()
    def test_reference_ids(self, sf):
        requests = [
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Id FROM Account LIMIT 1",
                "referenceId": "one",
            },
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
                "referenceId": "two",
            },
            {
                "method": "GET",
                "url": "/services/data/v50.0/query?q=SELECT Name FROM Account LIMIT 1",
                "referenceId": "three",
            },
        ]
        with CompositeParallelSalesforce(sf, 2, max_workers=1) as cpsf:
            results = cpsf.do_composite_requests(requests)
        assert len(list(results)) == len(requests)
        assert set(result["referenceId"] for result in results) == set(
            request["referenceId"] for request in requests
        )

    @pytest.mark.vcr()
    def test_errors(self, sf):
        requests = [{"method": "GET", "url": "/services/data/v50.0/sobjects/Foo"}]
        with CompositeParallelSalesforce(sf, 4, max_workers=1) as cpsf:
            results = cpsf.do_composite_requests(requests)
        assert results[0]["httpStatusCode"] == 404, results[0]["httpStatusCode"]



from salesforce_bulk import SalesforceBulk

from cumulusci.core.exceptions import ConfigError
from cumulusci.core.tasks import BaseSalesforceTask
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection


class BaseSalesforceApiTask(BaseSalesforceTask):
    name = "BaseSalesforceApiTask"
    api_version = None

    def _init_task(self):
        super()._init_task()
        self.sf = self._init_api()
        self.bulk = self._init_bulk()
        self.tooling = self._init_api("tooling")
        self._init_class()

    def _init_api(self, base_url=None):
        rv = get_simple_salesforce_connection(
            self.project_config,
            self.org_config,
            api_version=self.api_version,
            base_url=base_url,
        )

        return rv

    def _init_bulk(self):
        version = self.api_version or self.project_config.project__package__api_version
        if not version:
            raise ConfigError("Cannot find Salesforce version")
        return SalesforceBulk(
            host=self.org_config.instance_url.replace("https://", "").rstrip("/"),
            sessionId=self.org_config.access_token,
            API_version=version,
        )

    def _init_class(self):
        pass

    def _get_tooling_object(self, obj_name):
        obj = getattr(self.tooling, obj_name)
        obj.base_url = obj.base_url.replace("/sobjects/", "/tooling/sobjects/")
        return obj



"""
This class is used by test_pageobjects
"""
from cumulusci.robotframework.pageobjects import BasePage, pageobject


@pageobject(page_type="Test", object_name="Bar__c")
class BarTestPage(BasePage):
    def bar_keyword_1(self, message):
        self.builtin.log(message)
        return f"bar keyword 1: {message}"

    def bar_keyword_2(self, message):
        self.builtin.log(message)
        return f"bar keyword 2: {message}"



"""Salesforce customizations of VCR

Hides session IDs and OrgIDs.
Generally hides all request and response headers to save space in our repo.

Records transactions if there is an org specified and does not if there is not.
"""

import re

from .pytest_sf_orgconnect import sf_pytest_orgname


def simplify_body(request_or_response_body):
    decoded = request_or_response_body.decode("utf-8")
    if "<sessionId>" in decoded:
        decoded = re.sub(
            r"<sessionId>.*</sessionId>",
            "<sessionId>**Elided**</sessionId>",
            decoded,
        )
    decoded = re.sub(r"001[\w\d]{15,18}", "001ACCOUNTID", decoded)
    return decoded.encode()


def sf_before_record_request(request):
    if request.body:
        request.body = simplify_body(request.body)
    request.uri = re.sub(
        r"//.*.my.salesforce.com", "//orgname.my.salesforce.com", request.uri
    )
    request.uri = re.sub(
        r"//.*\d+.*.salesforce.com/", "//orgname.my.salesforce.com/", request.uri
    )
    request.uri = re.sub(r"00D[\w\d]{15,18}", "Organization/ORGID", request.uri)

    request.headers = {"Request-Headers": "Elided"}

    return request


# junk_headers = ["Public-Key-Pins-Report-Only", ]
def sf_before_record_response(response):
    response["headers"] = {"Response-Headers": "Elided"}
    if response.get("body"):
        response["body"]["string"] = simplify_body(response["body"]["string"])
    return response


def vcr_config(request):
    "Fixture for configuring VCR"

    orgname = sf_pytest_orgname(request)

    # https://vcrpy.readthedocs.io/en/latest/usage.html#record-modes
    if orgname:
        record_mode = "once"
    else:
        record_mode = "none"

    return {
        "record_mode": record_mode,
        "decode_compressed_response": True,
        "before_record_response": sf_before_record_response,
        "before_record_request": sf_before_record_request,
        # this is redundant, but I guess its a from of
        # security in-depth
        "filter_headers": [
            "Authorization",
            "Cookie",
            "Public-Key-Pins-Report-Only",
            "Last-Modified",
        ],
    }


replacements = [
    (re.compile(r"/v?\d\d.0/"), r"/vxx.0/"),
    (re.compile(r"/00D[\w\d]{10,20}"), "/ORGID"),
    (re.compile(r".com//"), r".com/"),
    (re.compile(r"ersion>\d\d.0<"), r"ersion>vxx.0<"),
]


def _noversion(s):
    if s:
        s = str(s, "utf-8") if isinstance(s, bytes) else s
        for pattern, replacement in replacements:
            s = pattern.sub(replacement, s)
        return s


def explain_mismatch(r1, r2):
    for a, b in zip(r1, r2):
        if a != b:
            print("MISMATCH\n\t", a, "\n!=\n\t", b)
    assert False
    return False


def salesforce_matcher(r1, r2):
    summary1 = (r1.method, _noversion(r1.uri), _noversion(r1.body))
    summary2 = (r2.method, _noversion(r2.uri), _noversion(r2.body))
    # uncomment explain_mismatch if you need to debug.
    # otherwise it will generate a lot of noise, even when things
    # are working properlly
    assert summary1 == summary2  # or explain_mismatch(summary1, summary2)


def salesforce_vcr(vcr):
    vcr.register_matcher("Salesforce Matcher", salesforce_matcher)
    vcr.match_on = ["Salesforce Matcher"]
    return vcr


salesforce_vcr.__doc__ = __doc__



import copy

from cumulusci.robotframework import locators_49

lex_locators = copy.deepcopy(locators_49.lex_locators)

lex_locators["actions"] = (
    "//runtime_platform_actions-actions-ribbon//ul"
    "|"
    "//ul[contains(concat(' ',normalize-space(@class),' '),' oneActionsRibbon ')]"
)

lex_locators["object"][
    "button"
] = "//div[contains(@class, 'slds-page-header')]//*[self::a[@title='{title}'] or self::button[@name='{title}']]"

lex_locators["record"]["header"][
    "field_value_link"
] = "//records-lwc-highlights-panel//force-highlights-details-item[.//*[.='{}']]//a"


lex_locators["record"]["related"] = {
    "button": "//*[@data-component-id='force_relatedListContainer']//article[.//span[@title='{}']]//a[@title='{}']",
    "card": "//*[@data-component-id='force_relatedListContainer']//article[.//span[@title='{}']]",
    "count": "//*[@data-component-id='force_relatedListContainer']//article//span[@title='{}']/following-sibling::span",
    "link": "//*[@data-component-id='force_relatedListContainer']//article[.//span[@title='{}']]//*[text()='{}']",
    "popup_trigger": "//*[@data-component-id='force_relatedListContainer']//article[.//span[@title='{}']]//span[text()='Show Actions']",
}



import itertools
import typing as T


def iterate_in_chunks(n: int, iterable: T.Iterable) -> T.Iterable[T.Tuple]:
    """Split an iterable into chunks of size 'n'."""
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk



def get_all_subclasses(cls):
    """Return all subclasses of the given class"""
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in get_all_subclasses(c)]
    )



from typing import Dict
from urllib.parse import urlparse

from cumulusci.core.config.oauth2_service_config import OAuth2ServiceConfig
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.oauth.client import OAuth2Client


class MarketingCloudServiceConfig(OAuth2ServiceConfig):
    def __init__(self, config, name, keychain):
        super().__init__(config, name, keychain)
        self._name = name
        self._keychain = keychain

    @classmethod
    def connect(cls, keychain: BaseProjectKeychain, kwargs: Dict):
        """This is called when a service is connected via `cci service connect`

        @param keychain - A keychain for accessing services
        @param kwargs - Any keyword arguments passed to `cci service connect`
        """
        client_config = keychain.get_service("oauth2_client", kwargs["oauth2_client"])
        config = client_config.config.copy()
        config["redirect_uri"] = config["callback_url"]
        oauth2_client = OAuth2Client(config)
        return oauth2_client.auth_code_flow()

    @property
    def tssd(self):
        """A dynamic value that represents the end user's subdomain.
        We can derive this value from either soap_instance_url or rest_instance_url
        which are present upon successful completion of an OAuth2 flow."""
        result = urlparse(self.config["rest_instance_url"])
        return result.netloc.split(".")[0]

    @property
    def access_token(self):
        return self._refresh_token()

    def _refresh_token(self):
        """Gets a fresh access token for the user.

        The info dict passed back from the OAuth flow includes:
        * access_token
        * refresh_token
        """
        oauth2_client_config = self._keychain.get_service(
            "oauth2_client", self.oauth2_client
        )
        oauth2_client = OAuth2Client(oauth2_client_config)
        info = oauth2_client.refresh_token(self.refresh_token)
        self.config.update(info)
        self.save()
        return info["access_token"]

    def save(self):
        assert self._keychain, "Keychain not set on MarketingCloudServiceConfig"
        self._keychain.set_service("marketing_cloud", self._name, self)



from json import JSONDecodeError
from unittest.mock import Mock, patch

from cumulusci import __version__
from cumulusci.core.config import OrgConfig
from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection


def test_connection():
    org_config = OrgConfig(
        {
            "instance_url": "https://orgname.my.salesforce.com",
            "access_token": "BOGUS",
        },
        "test",
    )
    proj_config = Mock()
    service_mock = Mock()
    service_mock.client_id = "TEST"
    proj_config.keychain.get_service.return_value = service_mock
    proj_config.project__package__api_version = "51.0"

    sf = get_simple_salesforce_connection(proj_config, org_config)
    assert sf.base_url == "https://orgname.my.salesforce.com/services/data/v51.0/"
    assert sf.headers["Authorization"] == "Bearer BOGUS"
    assert sf.headers["Sforce-Call-Options"] == "client=TEST"


def test_connection__explicit_api_version():
    org_config = OrgConfig(
        {
            "instance_url": "https://orgname.my.salesforce.com",
            "access_token": "BOGUS",
        },
        "test",
    )
    proj_config = Mock()
    service_mock = Mock()
    service_mock.client_id = "TEST"
    proj_config.keychain.get_service.return_value = service_mock

    sf = get_simple_salesforce_connection(proj_config, org_config, api_version="42.0")
    assert sf.base_url == "https://orgname.my.salesforce.com/services/data/v42.0/"


def test_connection__no_connected_app():
    org_config = OrgConfig(
        {
            "instance_url": "https://orgname.my.salesforce.com",
            "access_token": "BOGUS",
        },
        "test",
    )
    proj_config = Mock()
    proj_config.keychain.get_service.side_effect = ServiceNotConfigured

    sf = get_simple_salesforce_connection(proj_config, org_config)
    assert sf.headers["Sforce-Call-Options"] == f"client=CumulusCI/{__version__}"


def test_connection__with_port():
    org_config = OrgConfig(
        {
            "instance_url": "https://orgname.my.salesforce.com:8080",
            "access_token": "BOGUS",
        },
        "test",
    )
    proj_config = Mock()
    service_mock = Mock()
    service_mock.client_id = "TEST"
    proj_config.keychain.get_service.return_value = service_mock
    proj_config.project__package__api_version = "51.0"

    sf = get_simple_salesforce_connection(proj_config, org_config)
    assert sf.base_url == "https://orgname.my.salesforce.com:8080/services/data/v51.0/"


def test_sf_api_retries(mock_http_response):
    org_config = Mock()
    proj_config = Mock()
    service_mock = Mock()
    service_mock.client_id = "TEST"
    proj_config.keychain.get_service.return_value = service_mock
    org_config.instance_url = "https://enterprise-dream-6536.cs41.my.salesforce.com"
    org_config.access_token = "httpsenterprise-dream-6536.cs41.my.salesforce.com"

    sf = get_simple_salesforce_connection(proj_config, org_config, api_version="42.0")
    adapter = sf.session.get_adapter("http://")

    assert 0.3 == adapter.max_retries.backoff_factor
    assert 502 in adapter.max_retries.status_forcelist

    with patch(
        "urllib3.connectionpool.HTTPConnectionPool._make_request"
    ) as _make_request:
        _make_request.side_effect = [
            ConnectionResetError,
            mock_http_response(status=200),
        ]

        try:
            sf.describe()
        except JSONDecodeError:
            # We're not returning a message to decode
            pass

        assert 2 == _make_request.call_count



""" CLI logger """
import logging
import os
import sys
import tempfile

import coloredlogs
import requests

try:
    import colorama
except ImportError:
    # coloredlogs only installs colorama on Windows
    pass


def init_logger(log_requests=False):
    """Initialize the logger"""

    logger = logging.getLogger(__name__.split(".")[0])
    for handler in logger.handlers:  # pragma: no cover
        logger.removeHandler(handler)

    if os.name == "nt" and "colorama" in sys.modules:  # pragma: no cover
        colorama.init()

    formatter = coloredlogs.ColoredFormatter(fmt="%(asctime)s: %(message)s")
    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    if log_requests:
        requests.packages.urllib3.add_stderr_logger()


def get_tempfile_logger():
    """Creates a logger that writes to a temporary
    logfile. Returns the logger and path to tempfile"""
    logger = logging.getLogger("tempfile_logger")
    file_handle, filepath = tempfile.mkstemp()
    # close the file as it will be opened again by FileHandler
    os.close(file_handle)
    handler = logging.FileHandler(filepath, encoding="utf-8")
    handler.terminator = ""
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    return logger, filepath



import contextlib
import json
import pathlib
import textwrap
from typing import Optional

from cumulusci.core.utils import dictmerge
from cumulusci.tasks.metadata.package import PackageXmlGenerator
from cumulusci.tasks.salesforce import Deploy
from cumulusci.utils import temporary_dir

SETTINGS_XML = """<?xml version="1.0" encoding="UTF-8"?>
<{settingsName} xmlns="http://soap.sforce.com/2006/04/metadata">
{values}
</{settingsName}>"""
ORGPREF = """<preferences>
    <settingName>{name}</settingName>
    <settingValue>{value}</settingValue>
</preferences>"""


class DeployOrgSettings(Deploy):
    task_doc = """Deploys org settings from an sfdx scratch org definition file."""

    task_options = {
        "definition_file": {"description": "sfdx scratch org definition file"},
        "settings": {"description": "A dict of settings to apply"},
        "api_version": {"description": "API version used to deploy the settings"},
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        # We have no need for namespace injection when deploying settings,
        # so let's explicitly disable it to prevent the Deploy task
        # from making API calls to check if it's needed.
        self.options["managed"] = False
        self.options["namespaced_org"] = False

    def _run_task(self):
        settings = {}
        if self.options.get("definition_file"):
            with open(self.options["definition_file"], "r") as f:
                scratch_org_definition = json.load(f)
                settings = scratch_org_definition.get("settings", {})

        dictmerge(settings, self.options.get("settings", {}))

        if not settings:
            self.logger.info("No settings provided to deploy.")
            return

        api_version = (
            self.options.get("api_version") or self.org_config.latest_api_version
        )
        with build_settings_package(settings, None, api_version) as path:
            self.options["path"] = path
            return super()._run_task()


def capitalize(s):
    """
    Just capitalize first letter (different from .title, as it preserves
    the rest of the case).
    e.g. accountSettings -> AccountSettings
    """
    return s[0].upper() + s[1:]


@contextlib.contextmanager
def build_settings_package(
    settings: Optional[dict], object_settings: Optional[dict], api_version: str
):
    with temporary_dir() as path:
        if settings:
            pathlib.Path("settings").mkdir()
            for section, section_settings in settings.items():
                settings_name = capitalize(section)
                if section == "orgPreferenceSettings":
                    values = "    " + "\n    ".join(
                        line
                        for line in "\n".join(
                            ORGPREF.format(name=capitalize(k), value=v)
                            for k, v in section_settings.items()
                        ).splitlines()
                    )
                else:
                    values = textwrap.indent(_dict_to_xml(section_settings), "    ")
                # e.g. AccountSettings -> settings/Account.settings
                settings_file = (
                    pathlib.Path("settings")
                    / f"{settings_name[: -len('Settings')]}.settings"
                )
                with open(settings_file, "w", encoding="utf-8") as f:
                    f.write(
                        SETTINGS_XML.format(settingsName=settings_name, values=values)
                    )

        if object_settings:
            pathlib.Path("objects").mkdir()
            for obj_lower_name, this_obj_settings in object_settings.items():
                object_name = capitalize(obj_lower_name)
                file_content = _get_object_file(object_name, this_obj_settings)
                with open(
                    pathlib.Path("objects") / f"{object_name}.object",
                    "w",
                    encoding="utf-8",
                ) as f:
                    f.write(file_content)

        package_generator = PackageXmlGenerator(path, api_version)
        with open("package.xml", "w") as f:
            f.write(package_generator())

        yield path


def _get_object_file(object_name: str, settings: dict) -> str:
    sharing_model = ""
    if "sharingModel" in settings:
        sharing_model = (
            f"    <sharingModel>{capitalize(settings['sharingModel'])}</sharingModel>"
        )

    record_type = ""
    business_process = ""
    if "defaultRecordType" in settings:
        # Use the same default picklist values as SFDX to generate a Business Process,
        # if this object requires one.
        picklist_value = {
            "Case": "New",
            "Lead": "New - Not Contacted",
            "Opportunity": "Prospecting",
            "Solution": "Draft",
        }.get(object_name, "")
        default_status = (
            "<default>true</default>" if object_name == "Opportunity" else ""
        )

        business_process_reference = ""
        if picklist_value:
            # We need to add a Business Process
            business_process_reference = (
                f"<businessProcess>Default{object_name}</businessProcess>"
            )
            default_status = ""
            business_process = f"""    <businessProcesses>
        <fullName>Default{object_name}</fullName>
        <isActive>true</isActive>
        <values>
            <fullName>{picklist_value}</fullName>
            {default_status}
        </values>
    </businessProcesses>"""
        record_type = f"""    <recordTypes>
        <fullName>{capitalize(settings['defaultRecordType'])}</fullName>
        <label>{capitalize(settings['defaultRecordType'])}</label>
        <active>true</active>
        {business_process_reference}
    </recordTypes>
        """

    return f"""<?xml version="1.0" encoding="UTF-8"?>
<Object xmlns="http://soap.sforce.com/2006/04/metadata">
{sharing_model}
{record_type}
{business_process}
</Object>"""


def _dict_to_xml(d: dict) -> str:
    items = []
    for k, v in d.items():
        if isinstance(v, dict):
            v = "\n" + textwrap.indent(_dict_to_xml(v), "    ") + "\n"
        elif isinstance(v, str):
            pass
        elif isinstance(v, bool):
            v = str(v).lower()
        else:
            raise TypeError(f"Unexpected type {type(v)} for {k}")
        items.append(f"<{k}>{v}</{k}>")
    return "\n".join(items)



from pathlib import Path

import pytest

from cumulusci.tasks.preflight.packages import GetInstalledPackages


class TestIntegrationInfrastructure:
    "Test our two plugins for doing integration testing"

    @pytest.mark.vcr()
    def test_integration_tests(self, create_task):
        task = create_task(GetInstalledPackages, {})
        assert task() is not None

    def test_file_was_created(self):
        filename = (
            Path(__file__).parent
            / "cassettes/TestIntegrationInfrastructure.test_integration_tests.yaml"
        )

        assert filename.exists(), filename
        with filename.open() as f:
            data = f.read()
            assert "Bearer 0" not in data
            assert "Public-Key-Pins-Report-Only" not in data
            assert "<sessionId>00" not in data



import json
import re
import zipfile
from pathlib import Path
from unittest import mock

import pytest
import responses

from cumulusci.core.config import TaskConfig
from cumulusci.core.exceptions import DeploymentException
from cumulusci.tasks.marketing_cloud.deploy import (
    MCPM_ENDPOINT,
    MarketingCloudDeployTask,
)
from cumulusci.utils import temporary_dir


@pytest.fixture
def task(project_config):
    test_zip_file = Path(__file__).parent.absolute() / "test-mc-pkg.zip"
    task = MarketingCloudDeployTask(
        project_config,
        TaskConfig(
            {
                "options": {
                    "package_zip_file": test_zip_file.resolve(),
                    "custom_inputs": "companyName:Acme",
                }
            }
        ),
    )
    task.mc_config = mock.Mock()
    task.mc_config.access_token = "foo"
    task.mc_config.tssd = "bar"
    return task


@pytest.fixture
def task_without_custom_inputs(project_config):
    test_zip_file = Path(__file__).parent.absolute() / "test-mc-pkg.zip"
    task = MarketingCloudDeployTask(
        project_config,
        TaskConfig(
            {
                "options": {
                    "package_zip_file": test_zip_file.resolve(),
                    "custom_inputs": None,
                }
            }
        ),
    )
    task.mc_config = mock.Mock()
    task.mc_config.access_token = "foo"
    task.mc_config.tssd = "bar"
    return task


class TestMarketingCloudDeployTask:
    @responses.activate
    def test_run_task__deploy_succeeds_with_custom_inputs(self, task):
        responses.add(
            "POST",
            f"{MCPM_ENDPOINT}/deployments",
            json={"info": {"id": "JOBID", "status": "IN_PROGRESS"}},
        )
        responses.add(
            "GET",
            f"{MCPM_ENDPOINT}/deployments/JOBID",
            json={"status": "DONE", "entities": {}},
        )

        task.logger = mock.Mock()
        task._run_task()
        task.logger.info.assert_called_with("Deployment completed successfully.")
        assert task.logger.error.call_count == 0
        assert task.logger.warn.call_count == 0

    @responses.activate
    def test_run_task__deploy_succeeds_without_custom_inputs(
        self, task_without_custom_inputs
    ):
        responses.add(
            "POST",
            f"{MCPM_ENDPOINT}/deployments",
            json={"info": {"id": "JOBID", "status": "IN_PROGRESS"}},
        )
        responses.add(
            "GET",
            f"{MCPM_ENDPOINT}/deployments/JOBID",
            json={"status": "DONE", "entities": {}},
        )
        task = task_without_custom_inputs
        task.logger = mock.Mock()
        task._run_task()
        task.logger.info.assert_called_with("Deployment completed successfully.")
        assert task.logger.error.call_count == 0
        assert task.logger.warn.call_count == 0

    @responses.activate
    def test_run_task__deploy_fails(self, task):
        responses.add(
            "POST",
            f"{MCPM_ENDPOINT}/deployments",
            json={"info": {"id": "JOBID", "status": "IN_PROGRESS"}},
        )
        responses.add(
            "GET",
            f"{MCPM_ENDPOINT}/deployments/JOBID",
            json={
                "status": "DONE",
                "entities": {
                    "assets": {
                        0: {"status": "FAILED", "issues": ["A problem occurred"]},
                        1: {"status": "SKIPPED", "issues": ["A problem occurred"]},
                    },
                    "categories": {},
                },
            },
        )
        task.logger = mock.Mock()
        with pytest.raises(DeploymentException):
            task._run_task()

        assert task.logger.error.call_count == 2
        assert (
            task.logger.error.call_args_list[0][0][0]
            == "Failed to deploy assets/0. Status: FAILED. Issues: ['A problem occurred']"
        )
        assert (
            task.logger.error.call_args_list[1][0][0]
            == "Failed to deploy assets/1. Status: SKIPPED. Issues: ['A problem occurred']"
        )

    def test_zipfile_not_valid(self, task):
        task.options["package_zip_file"] = "not-a-valid-file.zip"
        task.logger = mock.Mock()
        task._run_task()

        task.logger.error.assert_called_once_with(
            "Package zip file not valid: not-a-valid-file.zip"
        )

    def test_construct_payload(self, task):
        pkg_zip_file = Path(task.options["package_zip_file"])
        with temporary_dir() as temp_dir:
            with zipfile.ZipFile(pkg_zip_file) as zf:
                zf.extractall(temp_dir)
                actual_payload = task._construct_payload(Path(temp_dir))

        expected_payload_file = (
            Path(__file__).parent.absolute() / "expected-payload.json"
        )
        with open(expected_payload_file, "r") as f:
            expected_payload = json.load(f)

        assert expected_payload == actual_payload

    def test_add_custom_inputs_to_payload__deployment_exception(self, task):
        custom_inputs = {"foo": "bar"}
        payload = {"input": [{"key": "baz"}]}
        error_message = re.escape("Custom input of key foo not found in package.")
        with pytest.raises(DeploymentException, match=error_message):
            task._add_custom_inputs_to_payload(custom_inputs, payload)

    def test_add_custom_inputs_to_payload(self, task):
        custom_inputs = {"companyName": "Acme"}
        payload = {"input": [{"key": "companyName"}]}
        payload = task._add_custom_inputs_to_payload(custom_inputs, payload)
        assert payload == {"input": [{"key": "companyName", "value": "Acme"}]}



import unittest
from pathlib import Path
from unittest import mock

import robot.api.logger

from cumulusci.core.config import BaseProjectConfig, UniversalConfig
from cumulusci.core.exceptions import TaskNotFoundError
from cumulusci.core.tests.utils import MockLoggerMixin
from cumulusci.robotframework.CumulusCI import CumulusCI
from cumulusci.utils import temporary_dir


class TestCumulusCILibrary(MockLoggerMixin, unittest.TestCase):
    def setUp(self):
        self.universal_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            self.universal_config,
            {
                "project": {"name": "Test"},
                "tasks": {
                    "get_pwd": {
                        "class_path": "cumulusci.tasks.command.Command",
                        "options": {
                            "command": "pwd",
                        },
                    },
                },
                "sources": {
                    "example": {"path": "/tmp"},
                },
            },
        )
        self.cumulusci = CumulusCI()
        self.cumulusci._project_config = self.project_config
        self.cumulusci._org = mock.Mock()

        self._task_log_handler.reset()
        self.task_log = self._task_log_handler.messages

    def test_run_task(self):
        """Smoke test; can we run the command task?"""
        result = self.cumulusci.run_task("get_pwd")
        self.assertEqual(result["returncode"], 0)

    def test_run_task_robot_logger(self):
        """Verify that 'run task' uses the robot logger"""
        with mock.patch.object(self.cumulusci, "_run_task"):
            self.cumulusci.run_task("get_pwd")
            args, kwargs = self.cumulusci._run_task.call_args
            task = args[0]
            self.assertEqual(task.logger, robot.api.logger)

    def test_robot_logger_supports_warning(self):
        """Verify that 'run task' uses a logger that supports .warning()

        Python deprecated the logger method "warn" in favor of
        "warning". Robot didn't get the memo and has "warn" instead of
        "warning".  Since our tasks use "warning", this verifies that
        we've patched the robot logger before passing it to the task
        constructor.

        """
        with mock.patch.object(self.cumulusci, "_run_task"):
            self.cumulusci.run_task("get_pwd")
            args, kwargs = self.cumulusci._run_task.call_args
            task = args[0]
            assert hasattr(
                task.logger, "warning"
            ), "robot logger should have a warning method but doesn't"

    def test_run_task_class_robot_logger(self):
        """Verify that 'run task class' uses the robot logger"""
        with mock.patch.object(self.cumulusci, "_run_task"):
            self.cumulusci.run_task_class(
                "cumulusci.tasks.command.Command", command="ls -l"
            )
            args, kwargs = self.cumulusci._run_task.call_args
            task = args[0]
            self.assertEqual(task.logger, robot.api.logger)

    def test_run_unknown_task(self):
        with self.assertRaises(TaskNotFoundError):
            self.cumulusci.run_task("bogus")

    def test_cross_project_task(self):
        """Verify that the cross-project task runs with the project config of the task
        See W-8891667
        """
        with temporary_dir() as tmpdir:
            tmpdir = Path(tmpdir).resolve()
            cumulusci_yml_path = tmpdir / "cumulusci.yml"
            with open(cumulusci_yml_path, "w+") as cumulusci_yml:
                self.project_config.sources["example"] = {"path": tmpdir}
                cumulusci_yml.write(
                    """
                    tasks:
                        whatever:
                            class_path: cumulusci.tasks.command.Command
                            options:
                                command: pwd
                    """
                )

            with mock.patch.object(self.cumulusci, "_run_task"):
                self.cumulusci.run_task("example:whatever")

                args, kwargs = self.cumulusci._run_task.call_args
                self.assertEqual(len(args), 1)

                # make sure it's not using the current project config
                # for the task, and that the config it _is_ using is
                # rooted in the directory we created
                task = args[0]
                self.assertNotEqual(task.project_config, self.cumulusci.project_config)
                self.assertEqual(tmpdir, Path(task.project_config.repo_root))



import copy

from cumulusci.robotframework import locators_50

lex_locators = copy.deepcopy(locators_50.lex_locators)

lex_locators["modal"]["button"] = "//div[contains(@class,'uiModal')]//button[.='{}']"

lex_locators["modal"]["has_error"] = "css: div.forceFormPageError"

lex_locators["modal"][
    "review_alert"
] = '//a[@force-recordediterror_recordediterror and text()="{}"]'

lex_locators["modal"]["field_alert"] = "//div[contains(@class, 'forceFormPageError')]"

# I like the new markup I'm seeing in Spring '21!
lex_locators["object"]["field"] = "//lightning-input[.//label[text()='{}']]//input"

lex_locators["record"]["related"]["button"] = (
    # the old locator searched for an <a> element, but in spring '21
    # the buttons are sometimes <button> elements. To be liberal in what
    # we accept, we'll just find anything that has the exact text.
    "//*[@data-component-id='force_relatedListContainer']"
    "//article[contains(@class, 'slds-card slds-card_boundary')]"
    "[.//span[@title='{}']]//*[text()='{}']"
)



import os
import sys
from pathlib import Path
from unittest.mock import Mock, patch

from cumulusci.utils.logging import (
    get_gist_logger,
    get_rot_file_logger,
    strip_ansi_sequences,
    tee_stdout_stderr,
)


class TestUtilLogging:
    @patch("cumulusci.utils.logging.get_gist_logger")
    def test_tee_stdout_stderr(self, gist_logger):
        args = ["cci", "test"]
        logger = Mock(handlers=[Mock()])
        gist_logger.return_value = Mock()
        # Setup temp logfile
        tempfile = "tempfile.log"
        log_content = "This is the content for the temp log file. And '┌' is an annoying character for Windows."
        with open(tempfile, "w", encoding="utf-8") as f:
            f.write(log_content)

        expected_stdout_text = "This is expected stdout.\n"
        expected_stderr_text = "This is expected stderr.\n"
        with tee_stdout_stderr(args, logger, tempfile):
            sys.stdout.write(expected_stdout_text)
            sys.stderr.write(expected_stderr_text)

        assert gist_logger.called_once()
        assert logger.debug.call_count == 3
        assert logger.debug.call_args_list[0][0][0] == "cci test\n"
        assert logger.debug.call_args_list[1][0][0] == expected_stdout_text
        assert logger.debug.call_args_list[2][0][0] == expected_stderr_text
        # temp log file should be deleted
        assert not os.path.isfile(tempfile)

    @patch("cumulusci.utils.logging.Path.mkdir")
    @patch("cumulusci.utils.logging.Path.home")
    @patch("cumulusci.utils.logging.get_rot_file_logger")
    def test_get_gist_logger(self, file_logger, home, mkdir):
        home.return_value = Path("/Users/bob.ross")
        get_gist_logger()
        file_logger.assert_called_once_with(
            "stdout/stderr", Path("/Users/bob.ross/.cumulusci/logs/cci.log")
        )

    @patch("cumulusci.utils.logging.RotatingFileHandler")
    @patch("cumulusci.utils.logging.logging")
    def test_get_rot_file_logger(self, logging, rot_filehandler):
        logger_name = "The happy logger"
        path = "happy/logger/path"
        logger = get_rot_file_logger(logger_name, path)

        logging.getLogger.assert_called_once_with(logger_name)
        rot_filehandler.assert_called_once_with(path, backupCount=5, encoding="utf-8")
        logger.addHandler.assert_called_once()
        logger.setLevel.assert_called_once()

    def test_strip_ansi_sequences(self):
        ansi_str = "\033[31mGoodbye ANSI color sequences!\033[0m"
        plain_str = "This is [just a plain old string with some] [symbols]"

        ansi_string_result = strip_ansi_sequences(ansi_str)
        plain_string_result = strip_ansi_sequences(plain_str)

        assert ansi_string_result == "Goodbye ANSI color sequences!"
        assert plain_string_result == plain_str



# -*- coding: utf-8 -*-
import unittest

import responses

from cumulusci.tasks.salesforce import ListCommunities

from .util import create_task

task_options = {}


class test_ListCommunities(unittest.TestCase):
    @responses.activate
    def test_lists_community(self):
        cc_task = create_task(ListCommunities, task_options)
        communities_url = "{}/services/data/v46.0/connect/communities".format(
            cc_task.org_config.instance_url
        )

        community_id = "000000000000000000"
        community_name = "Test Community"
        community_url_prefix = "test"
        community2_id = "000000000000000001"
        community2_name = "Test Community Д Two"

        responses.add(
            method=responses.GET,
            url=communities_url,
            status=200,
            json={
                "communities": [
                    {
                        "allowChatterAccessWithoutLogin": "false",
                        "allowMembersToFlag": "false",
                        "description": "This is a test community",
                        "id": community_id,
                        "invitationsEnabled": "false",
                        "knowledgeableEnabled": "false",
                        "loginUrl": f"https://mydomain.force.com/{community_url_prefix}/s/login",
                        "memberVisibilityEnabled": "true",
                        "name": community_name,
                        "nicknameDisplayEnabled": "false",
                        "privateMessagesEnabled": "false",
                        "reputationEnabled": "false",
                        "sendWelcomeEmail": "true",
                        "siteAsContainerEnabled": "true",
                        "siteUrl": f"https://mydomain.force.com/{community_url_prefix}",
                        "status": "Live",
                        "templateName": "VF Template",
                        "url": f"/services/data/v46.0/connect/communities/{community_id}",
                        "urlPathPrefix": community_url_prefix,
                    },
                    {
                        "allowChatterAccessWithoutLogin": "false",
                        "allowMembersToFlag": "false",
                        "description": "This is a test community",
                        "id": community2_id,
                        "invitationsEnabled": "false",
                        "knowledgeableEnabled": "false",
                        "loginUrl": "https://mydomain.force.com/s/login",
                        "memberVisibilityEnabled": "true",
                        "name": community2_name,
                        "nicknameDisplayEnabled": "false",
                        "privateMessagesEnabled": "false",
                        "reputationEnabled": "false",
                        "sendWelcomeEmail": "true",
                        "siteAsContainerEnabled": "true",
                        "siteUrl": "https://mydomain.force.com/",
                        "status": "Live",
                        "templateName": "VF Template",
                        "url": f"/services/data/v46.0/connect/communities/{community2_id}",
                        "urlPathPrefix": None,
                    },
                ],
                "total": "2",
            },
        )

        cc_task()

        assert 1 == len(responses.calls)
        assert communities_url == responses.calls[0].request.url
        assert cc_task.return_values == [community_name, community2_name]



from typing import Optional

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.utils import process_bool_arg, process_list_arg
from cumulusci.tasks.metadata_etl import MetadataSingleEntityTransformTask
from cumulusci.utils.xml.metadata_tree import MetadataElement


class AddRelatedLists(MetadataSingleEntityTransformTask):
    entity = "Layout"
    task_options = {
        "related_list": {
            "description": "Name of the Related List to include",
            "required": True,
        },
        "fields": {
            "description": "Array of field API names to include in the related list",
            "required": False,
        },
        "exclude_buttons": {
            "description": "Array of button names to suppress from the related list"
        },
        "custom_buttons": {
            "description": "Array of button names to add to the related list"
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _transform_entity(
        self, metadata: MetadataElement, api_name: str
    ) -> Optional[MetadataElement]:
        related_list = self._inject_namespace(self.options["related_list"])
        existing_related_lists = metadata.findall(
            "relatedLists", relatedList=related_list
        )

        if existing_related_lists:
            return None

        self._create_new_related_list(metadata, api_name, related_list)

        return metadata

    def _create_new_related_list(self, metadata, api_name, related_list):
        self.logger.info(f"Adding Related List {related_list} to {api_name}")

        fields = [
            self._inject_namespace(f)
            for f in process_list_arg(self.options.get("fields", []))
        ]
        exclude_buttons = [
            self._inject_namespace(f)
            for f in process_list_arg(self.options.get("exclude_buttons", []))
        ]
        custom_buttons = [
            self._inject_namespace(f)
            for f in process_list_arg(self.options.get("custom_buttons", []))
        ]

        elem = metadata.append("relatedLists")

        for f in fields:
            elem.append("fields", text=f)

        for button in exclude_buttons:
            elem.append("excludeButtons", button)

        for button in custom_buttons:
            elem.append("customButtons", button)
        elem.append("relatedList", text=related_list)


class AddFieldsToPageLayout(MetadataSingleEntityTransformTask):
    task_docs = """
        Inserts the listed fields or Visualforce pages into page layouts
        specified by API name.

        If the targeted item already exists, the layout metadata is not modified.

        You may supply a single position option, or multiple options for both pages and
        fields. The first option to to be matched will be used.

        Task option details:

        - fields:

            - api_name: [field API name]
            - required: Boolean (default False)
            - read_only: Boolean (default False, not compatable with required)
            - position: (Optional: A list of single or multiple position options.)

                - relative: [before | after | top | bottom]
                - field: [api_name] (Use with relative: before, after)
                - section: [index] (Use with relative: top, bottom)
                - column: [first | last] (Use with relative: top, bottom)

        - pages:

            - visualforce_page_name:[page_name]
            - height: int (Optional. Default: 200)
            - show_label: Boolean (Optional. Default: False)
            - show_scrollbars: Boolean (Optional. Default: False)
            - width: 0-100% (Optional. Default: 100%)
            - position: (Optional: A list of single or multiple position options.)

                - relative: [before | after | top | bottom]
                - field: [api_name] (Use with relative: before, after)
                - section: [index] (Use with relative: top, bottom)
                - column: [first | last] (Use with relative: top, bottom)
        """
    entity = "Layout"
    task_options = {
        "fields": {
            "description": "List of fields. See task info for structure.",
            "required": False,
        },
        "pages": {
            "description": "List of Visualforce Pages. See task info for structure.",
            "required": False,
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        self._adding_fields = process_list_arg(self.options.get("fields", []))
        self._adding_pages = process_list_arg(self.options.get("pages", []))
        # Split because a page could share an api name with a field in the same namespace
        self._existing_field_names = []
        self._existing_page_names = []

    def _transform_entity(
        self, metadata: MetadataElement, api_name: str
    ) -> Optional[MetadataElement]:

        self._metadata = metadata

        self._collect_existing_items()

        self._add_fields(api_name)
        self._add_pages(api_name)

        return self._metadata

    ## Collect fields already in the layout
    def _collect_existing_items(self):
        ## Need to traverse all layoutSections->layoutColumns->layoutItems->fields to find if item/field already exists
        layout_sections = self._metadata.findall("layoutSections")
        for section in layout_sections:
            layout_columns = section.findall("layoutColumns")
            for col in layout_columns:
                layout_items = col.findall("layoutItems")
                for item in layout_items:
                    if item.find("page"):
                        self._existing_page_names.append(item.page.text)
                    elif item.find("field"):
                        self._existing_field_names.append(item.field.text)

    def _add_fields(self, api_name):
        # Do nothing if there are no fields
        if not self._adding_fields:
            return

        field_props = {
            self._inject_namespace(field["api_name"]): field
            for field in self._adding_fields
        }

        layout_item_dict = self._add_items(
            self._adding_fields, self._existing_field_names, api_name
        )

        for field_item_key in layout_item_dict.keys():
            field = field_props.get(field_item_key)
            field_layout_item = layout_item_dict.get(field_item_key)
            adding_field_name = self._inject_namespace(field.get("api_name"))

            self.logger.info(f"Adding {adding_field_name} to {api_name}")

            behavior = "Edit"
            required = process_bool_arg(field.get("required") or False)
            read_only = process_bool_arg(field.get("read_only") or False)
            if required and not read_only:
                behavior = "Required"
            elif read_only and not required:
                behavior = "Readonly"
            elif required and read_only:
                raise TaskOptionsError(
                    f"`required` and `read_only` cannot both be True for {adding_field_name}"
                )
            field_layout_item.append("behavior", behavior)
            field_layout_item.append("field", adding_field_name)

    def _add_pages(self, api_name):
        # Do nothing if there are no fields
        if not self._adding_pages:
            return

        page_props = {
            self._inject_namespace(page["api_name"]): page
            for page in self._adding_pages
        }

        layout_item_dict = self._add_items(
            self._adding_pages, self._existing_page_names, api_name
        )

        for page_item_key in layout_item_dict.keys():
            page = page_props.get(page_item_key)
            page_layout_item = layout_item_dict.get(page_item_key)
            adding_page_name = self._inject_namespace(page.get("api_name"))

            self.logger.info(f"Adding {adding_page_name} to {api_name}")

            page_layout_item.append("page", adding_page_name)
            page_layout_item.append("height", str(page.get("height", 200)))
            page_layout_item.append(
                "showLabel",
                str(process_bool_arg(page.get("show_label") or False)).lower(),
            )
            page_layout_item.append(
                "showScrollbars",
                str(process_bool_arg(page.get("show_scrollbars") or False)).lower(),
            )
            page_layout_item.append("width", page.get("width", "100%"))

    def _add_items(self, item_list, existing_name_list, api_name):
        """Iterate over a list of new layout items that will return a dict{api_name:MetadataElement}"""

        item_dict = {}

        for item in item_list:
            item_name = self._inject_namespace(item.get("api_name"))
            if item_name in existing_name_list:
                self.logger.warning(
                    f"Skipped {item_name} because {item_name} is already present in {api_name}"
                )
                continue

            item_dict[item_name] = self._position_item(item)

        return item_dict

    def _position_item(self, new_item):
        """Returns new MetadataElement being added to the layout based on the positioning properties"""
        new_item_name = new_item.get("api_name")
        # Default to top of first section (index zero) of last column
        default = {
            "relative": "top",
            "section": 0,
            "column": "last",
        }

        # Position is optional
        position = new_item.get("position")

        if position is None:
            self.logger.warning(
                f"Position details are missing for: {new_item_name}. Default position is being applied."
            )
            position = [default]
        else:
            # append the default position to the list as a fallback
            position.append(default)

        for pos in position:
            new_layout_item = self._process_position(pos, default)
            if new_layout_item is not None:
                return new_layout_item

    def _process_position(self, position, default_position):
        """Attempt to apply a position specifier and return a MetadataElement representing a layout item, or None if the position specifier cannot be applied."""
        relative_position = position.get("relative", default_position.get("relative"))

        # Mutually exclusive, therefore the higher priority is field if specified
        relative_field_name = None
        if "field" in position:
            relative_field_name = (
                self._inject_namespace(position.get("field"))
                if position.get("field") is not None
                else None
            )
            relative_section_index = None
            if relative_position in ("top", "bottom"):
                # Missing a relative position to the field, therefore default to after
                relative_position = "after"
        elif "section" in position:
            relative_section_index = position.get("section")
        else:
            relative_section_index = default_position.get("section")
            relative_position = default_position.get("position")

        # Field relative
        if relative_field_name is not None and relative_position in (
            "before",
            "after",
        ):
            new_layout_item = self._new_layout_item(
                relative_field_name, relative_position
            )

        # Section relative
        if relative_section_index is not None and relative_position in (
            "top",
            "bottom",
        ):
            new_layout_item = self._new_section_item(
                relative_section_index,
                relative_position,
                position.get("column", default_position.get("column")),
            )

        return new_layout_item

    def _new_layout_item(self, field_text, position):
        for section in self._metadata.findall("layoutSections"):
            for col in section.findall("layoutColumns"):
                for item in col.findall("layoutItems"):
                    for field in item.findall("field"):
                        if field.text == field_text:
                            return getattr(col, "insert_" + position)(
                                item, "layoutItems"
                            )

    def _new_section_item(self, index, position, column_index):
        # Get section at index
        sections = list(self._metadata.findall("layoutSections"))

        if index > (len(sections) - 1):
            self.logger.warning(
                f"Unable to find section at index: {str(index)}. Default position is being selected."
            )
            return None

        section = sections[index]
        columns = list(section.findall("layoutColumns"))
        column = columns[-1 if column_index == "last" else 0]
        items = list(column.findall("layoutItems"))
        # Sections can be empty
        if len(items) == 0 or position == "bottom":
            return column.append("layoutItems")
        else:
            item_relative_position = "before" if position == "top" else "after"
            return getattr(column, "insert_" + item_relative_position)(
                items[0], "layoutItems"
            )


class AddRecordPlatformActionListItem(MetadataSingleEntityTransformTask):
    task_docs = """
        Inserts the targeted lightning button/action into specified
        layout's PlatformActionList with a 'Record' actionListContext.
        - If the targeted lightning button/action already exists,
            the layout metadata is not modified.
        - If there is no 'Record' context PlatformActionList,
            we will generate one and add the specified action

        Task definition example:

            dev_inject_apply_quick_action_into_account_layout:
            group: "Demo config and storytelling"
            description: Adds an Apply Quick Action button to the beggining of the button list on the Experiences Account Layout.
            class_path: tasks.layouts.InsertRecordPlatformActionListItem
            options:
                api_names: "Account-%%%NAMESPACE%%%Experiences Account Layout"
                action_name: "Account.Apply"
                action_type: QuickAction
                place_first: True

        Reference Documentation:
        https://developer.salesforce.com/docs/atlas.en-us.api_meta.meta/api_meta/meta_layouts.htm#PlatformActionList
        """

    entity = "Layout"
    task_options = {
        "action_type": {
            "description": "platformActionListItems.actionType like 'QuickAction' or 'CustomButton'",
            "required": True,
        },
        "action_name": {
            "description": "platformActionListItems.actionName. The API name for the action to be added.",
            "required": True,
        },
        "place_first": {
            "description": "When 'True' the specified Record platformActionListItem will be inserted before any existing on the layout. Default is 'False'",
            "required": False,
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        self._action_type = self.options.get("action_type", "")
        self._action_name = self._inject_namespace(self.options.get("action_name", ""))
        # Default to False if `place_first` option was not set
        self._place_first = process_bool_arg(self.options.get("place_first") or False)

    def _transform_entity(
        self, metadata: MetadataElement, api_name: str
    ) -> Optional[MetadataElement]:
        # get or create an existing action list
        self._existing_action_list = self._get_existing_action_list(metadata)

        # check for existing Platform Action List Item of same name
        #   (i.e. The desired QuickAction is in this layout already)
        if self._existing_action_list.find(
            "platformActionListItems", actionName=self._action_name
        ):
            self.logger.info(
                f"Action named {self._action_name} already exists in {api_name}, task exiting without modifying layout."
            )
            return None

        # create an action list item
        self._create_new_action_list_item(self._existing_action_list)
        self._update_platform_action_list_items_sort_order(self._existing_action_list)
        return metadata

    def _get_existing_action_list(self, metadata: MetadataElement):
        existing_action_list = metadata.find(
            "platformActionList", actionListContext="Record"
        )
        if not existing_action_list:
            existing_action_list = metadata.append("platformActionList")
            existing_action_list.append("actionListContext", "Record")

        return existing_action_list

    def _create_new_action_list_item(self, existing_action_list):
        if self._place_first and existing_action_list.find("platformActionListItems"):
            existing_action_list.insert_after(
                existing_action_list.find("actionListContext"),
                "platformActionListItems",
            )
            action_list_item = existing_action_list.find("platformActionListItems")
        else:
            action_list_item = existing_action_list.append("platformActionListItems")

        action_list_item.append("actionName", self._action_name)
        action_list_item.append("actionType", self._action_type)
        action_list_item.append("sortOrder", "place_first:" + str(self._place_first))

    def _update_platform_action_list_items_sort_order(self, existing_action_list):
        """
        Updates the sortOrder element of the platformActionListItems
            Takes a platformActionList (MetadatataElement) and sets the sortOrder according
            to platformActionListItem placement relative to siblings
        """
        for index, child in enumerate(
            existing_action_list.findall("platformActionListItems")
        ):
            sortOrder = child.find("sortOrder")
            sortOrder.text = str(index)



import gzip
from collections import defaultdict
from contextlib import ExitStack, contextmanager
from email.utils import parsedate
from logging import getLogger
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Iterable, List, NamedTuple, Optional, Tuple

from sqlalchemy import MetaData, create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import create_session, exc, sessionmaker

from cumulusci.salesforce_api.org_schema_models import (
    Base,
    Field,
    FileMetadata,
    SObject,
)
from cumulusci.utils.http.multi_request import CompositeParallelSalesforce

y2k = "Sat, 1 Jan 2000 00:00:01 GMT"


def zip_database(tempfile, schema_path):
    """Compress tempfile.db to schema_path.db.gz"""
    with tempfile.open("rb") as db:
        with schema_path.open("wb") as fileobj:
            with gzip.GzipFile(fileobj=fileobj, mode="w") as gzipped:
                gzipped.write(db.read())


def unzip_database(gzipfile, outfile):
    """Decompress schema_path.db.gz to outfile.db"""
    with gzipfile.open("rb") as fileobj:
        with gzip.GzipFile(fileobj=fileobj) as gzipped:
            with open(outfile, "wb") as db:
                db.write(gzipped.read())


class Schema:
    """Represents an org's schema, cached from describe() calls"""

    _last_modified_date = None

    def __init__(self, engine, schema_path):
        self.engine = engine
        Session = sessionmaker(bind=self.engine)
        self.session = Session()
        self.path = schema_path

    @property
    def sobjects(self):
        return self.session.query(SObject)

    def __getitem__(self, name):
        try:
            return self.session.query(SObject).filter_by(name=name).one()
        except exc.NoResultFound:
            raise KeyError(f"No sobject named {name}")

    def __contains__(self, name):
        return self.session.query(SObject).filter_by(name=name).all()

    def keys(self):
        return (x[0] for x in self.session.query(SObject.name).all())

    def values(self):
        return self.session.query(SObject).all()

    def items(self):
        return ((obj.name, obj) for obj in self.session.query(SObject).all())

    def block_writing(self):
        """After this method is called, the database can't be updated again"""
        # changes don't get saved back to the gzip
        # so there is no point writing to the DB
        def closed():
            raise IOError("Database is not open for writing")

        self.session._real_commit__ = self.session.commit
        self.session.commit = closed

    def close(self):
        self.session.close()

    @property
    def last_modified_date(self):
        """Date of the most recent schema update"""
        if not self._last_modified_date:
            try:
                self._last_modified_date = (
                    self.session.query(FileMetadata)
                    .filter(FileMetadata.name == "Last-Modified")
                    .one()
                    .value
                )
            except exc.NoResultFound:
                pass
        return self._last_modified_date

    def __repr__(self):
        return f"<Schema {self.path} : {self.engine}>"

    def populate_cache(self, sf, last_modified_date, logger=None):
        """Populate a schema cache from the API, using last_modified_date
        to pull down only new schema"""

        sobjs = sf.describe()["sobjects"]
        sobj_names = [obj["name"] for obj in sobjs]

        responses = list(deep_describe(sf, last_modified_date, sobj_names))
        changes = [
            (resp.body, resp.last_modified_date)
            for resp in responses
            if resp.status == 200
        ]
        unexpected = [resp for resp in responses if resp.status not in (200, 304)]
        for unknown in unexpected:
            logger.warning(
                f"Unexpected describe reply. An SObject may be missing: {unknown}"
            )
        self._populate_cache_from_describe(changes, last_modified_date)

    def _populate_cache_from_describe(
        self, describe_objs: List[Tuple[dict, str]], last_modified_date
    ):
        """Populate a schema cache from a list of describe objects."""
        engine = self.engine
        metadata = Base.metadata
        metadata.bind = engine
        metadata.reflect()

        with BufferedSession(engine, metadata) as sess:

            max_last_modified = (parsedate(last_modified_date), last_modified_date)
            for (sobj_data, last_modified) in describe_objs:
                fields = sobj_data.pop("fields")
                create_row(sess, SObject, sobj_data)
                for field in fields:
                    field["sobject"] = sobj_data["name"]
                    create_row(sess, Field, field)
                    sortable = parsedate(last_modified), last_modified
                    if sortable > max_last_modified:
                        max_last_modified = sortable

            create_row(
                sess,
                FileMetadata,
                {"name": "Last-Modified", "value": max_last_modified[1]},
            )
            create_row(sess, FileMetadata, {"name": "FormatVersion", "value": 1})

        engine.execute("vacuum")


def create_row(buffered_session: "BufferedSession", model, valuesdict: dict):
    buffered_session.write_single_row(model.__tablename__, valuesdict)


class BufferedSession:
    """Buffer writes to a SQL DB for faster performmance"""

    def __init__(self, engine: Engine, metadata: MetaData, max_buffer_size: int = 1000):
        self.buffered_rows = defaultdict(list)
        self.columns = {}
        self.engine = engine
        self.metadata = metadata
        self._prepare()
        self.max_buffer_size = max_buffer_size

    def __enter__(self, *args):
        self.session = create_session(bind=self.engine, autocommit=False)
        return self

    def __exit__(self, *args):
        self.close()

    def _prepare(self):
        # Setup table info used by the write-buffering infrastructure
        assert self.metadata.tables
        self.insert_statements = {}
        for tablename, model in self.metadata.tables.items():
            self.insert_statements[tablename] = model.insert(bind=self.engine)
            self.columns[tablename] = {
                colname: None for colname in model.columns.keys()
            }

    def write_single_row(self, tablename: str, row: Dict) -> None:
        # but first, normalize it so all keys have a value. SQLite Requires it.
        normalized_row = {}
        for key, default in self.columns[tablename].items():
            normalized_row[key] = row.get(key, default)

        # cache the value for later insert
        self.buffered_rows[tablename].append(normalized_row)

        # flush if buffer is full
        if len(self.buffered_rows[tablename]) > self.max_buffer_size:
            self.flush()

    def flush(self):
        for tablename, insert_statement in self.insert_statements.items():
            # seems much more efficient than line-by-line inserts.
            if self.buffered_rows[tablename]:
                self.session.execute(insert_statement, self.buffered_rows[tablename])
                self.buffered_rows[tablename] = []
        self.session.flush()

    def commit(self):
        self.flush()
        self.session.commit()

    def close(self) -> None:
        self.commit()
        self.session.close()


@contextmanager
def get_org_schema(sf, org_config, force_recache=False, logger=None):
    """
    Get a read-only representation of an org's schema.

    org_config - an OrgConfig for the relevant org
    force_recache: True - replace cache. False (default) - use/update cache is available.
    logger - replace the standard logger "cumulusci.salesforce_api.org_schema"
    """
    with org_config.get_orginfo_cache_dir(Schema.__module__) as directory:
        directory.mkdir(exist_ok=True, parents=True)
        schema_path = directory / "org_schema.db.gz"

        if force_recache and schema_path.exists():
            schema_path.unlink()

        logger = logger or getLogger(__name__)

        with ExitStack() as closer:
            tempdir = TemporaryDirectory()
            closer.enter_context(tempdir)
            tempfile = Path(tempdir.name) / "temp_org_schema.db"
            schema = None
            if schema_path.exists():
                try:
                    cleanups_on_failure = []
                    unzip_database(schema_path, tempfile)
                    cleanups_on_failure.extend([schema_path.unlink, tempfile.unlink])
                    engine = create_engine(f"sqlite:///{str(tempfile)}")

                    schema = Schema(engine, schema_path)
                    cleanups_on_failure.append(schema.close)
                    closer.callback(schema.close)
                    assert schema.sobjects.first().name
                    schema.from_cache = True
                except Exception as e:
                    logger.warning(
                        f"Cannot read `{schema_path}`. Recreating it. Reason `{e}`."
                    )
                    schema = None
                    for cleanup_action in reversed(cleanups_on_failure):
                        cleanup_action()

            if schema is None:
                engine = create_engine(f"sqlite:///{str(tempfile)}")
                Base.metadata.bind = engine
                Base.metadata.create_all()
                schema = Schema(engine, schema_path)
                closer.callback(schema.close)
                schema.from_cache = False

            schema.populate_cache(
                sf,
                schema.last_modified_date or y2k,
                logger,
            )
            schema.block_writing()
            # save a gzipped copy for later
            zip_database(tempfile, schema_path)
            yield schema


class DescribeResponse(NamedTuple):
    """Result of a describe call from Salesforce"""

    status: int
    body: dict
    last_modified_date: str = None
    refId: str = None


def deep_describe(
    sf, last_modified_date: Optional[str], objs: List[str]
) -> Iterable[DescribeResponse]:
    """Fetch describe data for changed sobjects

    Fetch describe data for sobjects from the list 'objs'
    which have changed since last_modified_date (in HTTP
    proto format) and yield each object as a DescribeResponse object."""
    last_modified_date = last_modified_date or y2k
    with CompositeParallelSalesforce(sf, max_workers=8) as cpsf:
        responses = cpsf.do_composite_requests(
            (
                {
                    "method": "GET",
                    "url": f"/services/data/v{sf.sf_version}/sobjects/{obj}/describe",
                    "referenceId": f"ref{obj}",
                    "httpHeaders": {"If-Modified-Since": last_modified_date},
                }
                for obj in objs
            )
        )

        responses = (
            DescribeResponse(
                response["httpStatusCode"],
                response["body"],
                response["httpHeaders"].get("Last-Modified"),
                response["referenceId"],
            )
            for response in responses
        )
        yield from responses



import json
import random
import re
import string
from pathlib import Path

from cumulusci.cli.ui import CliTable
from cumulusci.core.exceptions import SalesforceException
from cumulusci.core.utils import process_bool_arg, process_list_arg
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.utils import inject_namespace

API_ROLLBACK_MESSAGE = "The transaction was rolled back since another operation in the same transaction failed."
API_INVALID_REF_MESSAGE = "Invalid reference specified."


class CompositeApi(BaseSalesforceApiTask):
    task_docs = """
This task is a wrapper for Composite REST API calls. Given a list of JSON files
(one request body per file), POST each and process the returned composite
result. Files are processed in the order given by the ``data_files`` option.

In addition, this task will process the request body and replace namespace
(``%%%NAMESPACE%%%``) and user ID (``%%%USERID%%%``) tokens. To avoid username
collisions, use the ``randomize_username`` option to replace the top-level
domains in any ``Username`` field with a random string.

When the top-level ``allOrNone`` property for the request is set to true a
SalesforceException is raised if an error is returned for any subrequest,
otherwise partial successes will not raise an exception.

Example Task Definition
-----------------------

.. code-block::  yaml

  tasks:
      example_composite_request:
          class_path: cumulusci.tasks.salesforce.composite.CompositeApi
          options:
             data_files:
                 - "datasets/composite/users.json"
                 - "datasets/composite/setup_objects.json"
    """

    task_options = {
        "data_files": {
            "description": "A list of paths, where each path is a JSON file containing a composite request body.",
            "required": True,
        },
        "managed": {
            "description": "If True, replaces namespace tokens with the namespace prefix.",
            "default": False,
        },
        "namespaced": {
            "description": "If True, replaces namespace tokens with the namespace prefix.",
            "default": False,
        },
        "randomize_username": {
            "description": "If True, randomize the TLD for any 'Username' fields.",
            "default": False,
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.data_files = process_list_arg(self.options.get("data_files") or [])

    def _run_task(self):
        for data_file_path in self.data_files:
            self.logger.info(f"Processing {data_file_path}")
            self._composite_request(data_file_path)

    def _composite_request(self, data_file_path):
        request_body = self._process_json(Path(data_file_path).read_text())
        self.is_all_or_none = json.loads(request_body).get("allOrNone", False)
        result = self.sf.restful("composite", method="POST", data=request_body)
        self._process_response(result)

    def _process_json(self, body):
        """Replace namespace tokens and randomize username domains."""
        user_id = self.org_config.user_id
        body = body.replace("%%%USERID%%%", user_id)

        namespace = self.project_config.project__package__namespace
        if "managed" in self.options:
            managed = process_bool_arg(self.options["managed"])
        else:
            managed = (
                bool(namespace) and namespace in self.org_config.installed_packages
            )

        _, body = inject_namespace(
            "composite",
            body,
            namespace=namespace,
            managed=managed,
            namespaced_org=self.options.get("namespaced", self.org_config.namespaced),
        )

        if self.options.get("randomize_username", False):
            random_tld = "".join(random.choices(string.ascii_lowercase, k=4))
            body = re.sub(
                r'("Username": .[\w-]+@[\w-]+\.)+[\w-]+', fr"\1{random_tld}", body
            )

        return body

    def _process_response(self, result):
        """Handle the compositeResponse and raise an exception if failed."""
        subrequests = result["compositeResponse"]
        status_codes = {subrequest["httpStatusCode"] for subrequest in subrequests}

        all_success = all([self._http_ok(code) for code in status_codes])
        if self.is_all_or_none and not all_success:
            self._log_exception_message(subrequests)
            raise SalesforceException(json.dumps(subrequests, indent=2))
        else:
            self._log_success_message(subrequests)

    def _http_ok(self, status_code):
        return status_code >= 200 and status_code < 300

    def _log_exception_message(self, subrequests):
        table_data = [["ReferenceId", "Message"]]
        table_data.extend(
            [sub["referenceId"], body["message"]]
            for sub in subrequests
            for body in sub["body"]
            if body.get("message") != API_ROLLBACK_MESSAGE
        )
        table = CliTable(table_data, wrap_cols=["Message"])
        self.logger.error("The request failed with the following message(s):\n\n")
        self.logger.error("\n" + str(table))

    def _log_success_message(self, subrequests):
        table_data = [["ReferenceId", "Success"]]
        table_data.extend(
            [sub["referenceId"], self._http_ok(sub["httpStatusCode"])]
            for sub in subrequests
        )
        table = CliTable(table_data, bool_cols=["Success"], title="Subrequest Results")
        self.logger.info("\n" + str(table))



from unittest import mock

import click
import pytest

from cumulusci.cli.runtime import CliRuntime
from cumulusci.core.config import FlowConfig
from cumulusci.core.exceptions import CumulusCIException, FlowNotFoundError
from cumulusci.core.flowrunner import FlowCoordinator

from .. import flow
from .utils import DummyTask, run_click_command


@mock.patch("cumulusci.cli.flow.CliTable")
def test_flow_list(cli_tbl):
    runtime = mock.Mock()
    runtime.get_available_flows.return_value = [
        {"name": "test_flow", "description": "Test Flow", "group": "Testing"}
    ]
    runtime.universal_config.cli__plain_output = None
    run_click_command(flow.flow_list, runtime=runtime, plain=False, print_json=False)

    cli_tbl.assert_called_with(
        [["Flow", "Description"], ["test_flow", "Test Flow"]],
        "Testing",
        wrap_cols=["Description"],
    )


@mock.patch("json.dumps")
def test_flow_list_json(json_):
    flows = [{"name": "test_flow", "description": "Test Flow"}]
    runtime = mock.Mock()
    runtime.get_available_flows.return_value = flows
    runtime.universal_config.cli__plain_output = None

    run_click_command(flow.flow_list, runtime=runtime, plain=False, print_json=True)

    json_.assert_called_with(flows)


@mock.patch("click.echo")
def test_flow_info(echo):
    runtime = mock.Mock()
    flow_config = FlowConfig({"description": "Test Flow", "steps": {}})
    runtime.get_flow.return_value = FlowCoordinator(None, flow_config)

    run_click_command(flow.flow_info, runtime=runtime, flow_name="test")

    echo.assert_called_with("Description: Test Flow")


def test_flow_info__not_found():
    runtime = mock.Mock()
    runtime.get_flow.side_effect = FlowNotFoundError
    with pytest.raises(click.UsageError):
        run_click_command(flow.flow_info, runtime=runtime, flow_name="test")


@mock.patch("cumulusci.cli.flow.group_items")
@mock.patch("cumulusci.cli.flow.document_flow")
def test_flow_doc__no_flows_rst_file(doc_flow, group_items):
    runtime = mock.Mock()
    runtime.universal_config.flows = {"test": {}}
    flow_config = FlowConfig({"description": "Test Flow", "steps": {}})
    runtime.get_flow.return_value = FlowCoordinator(None, flow_config)

    group_items.return_value = {"Group One": [["test flow", "description"]]}

    run_click_command(flow.flow_doc, runtime=runtime)
    group_items.assert_called_once()
    doc_flow.assert_called()


@mock.patch("click.echo")
@mock.patch("cumulusci.cli.flow.load_yaml_data")
def test_flow_doc__with_flows_rst_file(load_yaml_data, echo):
    runtime = CliRuntime(
        config={
            "flows": {
                "Flow1": {
                    "steps": {},
                    "description": "Description of Flow1",
                    "group": "Group1",
                }
            }
        },
        load_keychain=False,
    )

    load_yaml_data.return_value = {
        "intro_blurb": "opening blurb for flow reference doc",
        "groups": {
            "Group1": {"description": "This is a description of group1."},
        },
        "flows": {"Flow1": {"rst_text": "Some ``extra`` **pizzaz**!"}},
    }

    run_click_command(flow.flow_doc, runtime=runtime)

    assert 1 == load_yaml_data.call_count

    expected_call_args = [
        "Flow Reference\n==========================================\n\nopening blurb for flow reference doc\n\n",
        "Group1\n------",
        "This is a description of group1.",
        "Flow1\n^^^^^\n\n**Description:** Description of Flow1\n\nSome ``extra`` **pizzaz**!\n**Flow Steps**\n\n.. code-block:: console\n",
        "",
    ]
    expected_call_args = [mock.call(s) for s in expected_call_args]
    assert echo.call_args_list == expected_call_args


def test_flow_run():
    org_config = mock.Mock(scratch=True, config={})
    runtime = CliRuntime(
        config={
            "flows": {"test": {"steps": {1: {"task": "test_task"}}}},
            "tasks": {
                "test_task": {
                    "class_path": "cumulusci.cli.tests.test_flow.DummyTask",
                    "description": "Test Task",
                }
            },
        },
        load_keychain=False,
    )
    runtime.get_org = mock.Mock(return_value=("test", org_config))
    runtime.get_flow = mock.Mock()

    run_click_command(
        flow.flow_run,
        runtime=runtime,
        flow_name="test",
        org="test",
        delete_org=True,
        debug=False,
        o=[("test_task__color", "blue")],
        skip=(),
        no_prompt=True,
    )

    runtime.get_flow.assert_called_once_with(
        "test", options={"test_task": {"color": "blue"}}
    )
    org_config.delete_org.assert_called_once()


def test_flow_run__delete_org_when_error_occurs_in_flow():
    org_config = mock.Mock(scratch=True, config={})
    runtime = CliRuntime(
        config={
            "flows": {"test": {"steps": {1: {"task": "test_task"}}}},
            "tasks": {
                "test_task": {
                    "class_path": "cumulusci.cli.tests.test_flow.DummyTask",
                    "description": "Test Task",
                }
            },
        },
        load_keychain=False,
    )
    runtime.get_org = mock.Mock(return_value=("test", org_config))
    coordinator = mock.Mock()
    coordinator.run.side_effect = CumulusCIException
    runtime.get_flow = mock.Mock(return_value=coordinator)

    with pytest.raises(CumulusCIException):
        run_click_command(
            flow.flow_run,
            runtime=runtime,
            flow_name="test",
            org="test",
            delete_org=True,
            debug=False,
            o=[("test_task__color", "blue")],
            skip=(),
            no_prompt=True,
        )

    runtime.get_flow.assert_called_once_with(
        "test", options={"test_task": {"color": "blue"}}
    )
    org_config.delete_org.assert_called_once()


def test_flow_run__option_error():
    org_config = mock.Mock(scratch=True, config={})
    runtime = CliRuntime(config={"noop": {}}, load_keychain=False)
    runtime.get_org = mock.Mock(return_value=("test", org_config))

    with pytest.raises(click.UsageError, match="-o"):
        run_click_command(
            flow.flow_run,
            runtime=runtime,
            flow_name="test",
            org="test",
            delete_org=True,
            debug=False,
            o=[("test_task", "blue")],
            skip=(),
            no_prompt=True,
        )


def test_flow_run__delete_non_scratch():
    org_config = mock.Mock(scratch=False)
    runtime = mock.Mock()
    runtime.get_org.return_value = ("test", org_config)

    with pytest.raises(click.UsageError):
        run_click_command(
            flow.flow_run,
            runtime=runtime,
            flow_name="test",
            org="test",
            delete_org=True,
            debug=False,
            o=None,
            skip=(),
            no_prompt=True,
        )


@mock.patch("click.echo")
def test_flow_run__org_delete_error(echo):
    org_config = mock.Mock(scratch=True, config={})
    org_config.delete_org.side_effect = Exception
    org_config.save_if_changed.return_value.__enter__ = lambda *args: ...
    org_config.save_if_changed.return_value.__exit__ = lambda *args: ...
    runtime = CliRuntime(
        config={
            "flows": {"test": {"steps": {1: {"task": "test_task"}}}},
            "tasks": {
                "test_task": {
                    "class_path": "cumulusci.cli.tests.test_flow.DummyTask",
                    "description": "Test Task",
                }
            },
        },
        load_keychain=False,
    )
    runtime.get_org = mock.Mock(return_value=("test", org_config))
    DummyTask._run_task = mock.Mock()

    kwargs = {
        "runtime": runtime,
        "flow_name": "test",
        "org": "test",
        "delete_org": True,
        "debug": False,
        "no_prompt": True,
        "o": (("test_task__color", "blue"),),
        "skip": (),
    }

    run_click_command(flow.flow_run, **kwargs)

    echo.assert_any_call(
        "Scratch org deletion failed.  Ignoring the error below to complete the flow:"
    )



import tempfile
from pathlib import Path
from unittest import mock

import pytest
from lxml import etree

from cumulusci.core.exceptions import CumulusCIException, TaskOptionsError
from cumulusci.tasks.metadata_etl import (
    BaseMetadataETLTask,
    BaseMetadataSynthesisTask,
    BaseMetadataTransformTask,
    MetadataSingleEntityTransformTask,
    UpdateMetadataFirstChildTextTask,
)
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.utils.xml.metadata_tree import fromstring


class MetadataETLTask(BaseMetadataETLTask):
    _get_package_xml_content = mock.Mock()
    _transform = mock.Mock()


class TestBaseMetadataETLTask:
    def test_init_options(self):
        task = create_task(MetadataETLTask, {"managed": True, "api_version": "47.0"})

        assert task.options["managed"]
        assert task.options["api_version"] == "47.0"

    def test_init_options_raises_invalid_api_value(self):
        with pytest.raises(TaskOptionsError):
            create_task(MetadataETLTask, {"api_version": "aaa"})

    def test_inject_namespace(self):
        task = create_task(
            MetadataETLTask,
            {"managed": True, "namespace_inject": "test", "api_version": "47.0"},
        )

        assert task._inject_namespace("%%%NAMESPACE%%%Test__c") == "test__Test__c"
        task.options["managed"] = False
        assert task._inject_namespace("%%%NAMESPACE%%%Test__c") == "Test__c"

    @mock.patch("cumulusci.tasks.metadata_etl.base.ApiRetrieveUnpackaged")
    def test_retrieve(self, api_mock):
        task = create_task(
            MetadataETLTask,
            {"managed": False, "namespace_inject": "test", "api_version": "47.0"},
        )
        task.retrieve_dir = mock.Mock()
        task._get_package_xml_content = mock.Mock()
        task._get_package_xml_content.return_value = ""

        task._retrieve()
        api_mock.assert_called_once_with(
            task, task._generate_package_xml(False), "47.0"
        )
        api_mock.return_value.assert_called_once_with()
        api_mock.return_value.return_value.extractall.assert_called_once_with(
            task.retrieve_dir
        )

    @mock.patch("cumulusci.tasks.salesforce.Deploy")
    def test_deploy(self, deploy_mock):
        with tempfile.TemporaryDirectory() as tmpdir:
            task = create_task(
                MetadataETLTask,
                {"managed": False, "namespace_inject": "test", "api_version": "47.0"},
            )
            task.deploy_dir = Path(tmpdir)
            task._generate_package_xml = mock.Mock()
            task._generate_package_xml.return_value = "test"
            result = task._deploy()
            assert (Path(tmpdir) / "package.xml").read_text() == "test"

            assert len(deploy_mock.call_args_list) == 1

            assert deploy_mock.call_args_list[0][0][0] == task.project_config
            assert deploy_mock.call_args_list[0][0][2] == task.org_config
            deploy_mock.return_value.assert_called_once_with()
            assert result == deploy_mock.return_value.return_value

    def test_run_task(self):
        task = create_task(
            MetadataETLTask,
            {"managed": False, "namespace_inject": "test", "api_version": "47.0"},
        )

        task._retrieve = mock.Mock()
        task._deploy = mock.Mock()
        task.retrieve = True
        task.deploy = True

        task()

        task._retrieve.assert_called_once_with()
        task._deploy.assert_called_once_with()


class MetadataSynthesisTask(BaseMetadataSynthesisTask):
    _get_package_xml_content = mock.Mock()
    _synthesize = mock.Mock()


class TestBaseMetadataSynthesisTask:
    def test_synthesis(self):
        task = create_task(
            MetadataSynthesisTask,
            {"managed": False, "namespace_inject": "test", "api_version": "47.0"},
        )

        task._deploy = mock.Mock()
        task._synthesize()
        task._synthesize = mock.Mock()

        task()

        task._deploy.assert_called_once_with()
        task._synthesize.assert_called_once_with()

    @mock.patch("cumulusci.tasks.metadata_etl.base.PackageXmlGenerator")
    def test_generate_package_xml(self, package_mock):
        task = create_task(
            MetadataSynthesisTask,
            {"managed": False, "namespace_inject": "test", "api_version": "47.0"},
        )
        task.deploy_dir = "test"

        result = task._generate_package_xml(True)
        package_mock.assert_called_once_with(str(task.deploy_dir), task.api_version)
        package_mock.return_value.assert_called_once_with()
        assert result == package_mock.return_value.return_value


class MetadataTransformTask(BaseMetadataTransformTask):
    _get_entities = mock.Mock()
    _transform = mock.Mock()


class TestBaseMetadataTransformTask:
    def test_generate_package_xml(self):
        task = create_task(
            MetadataTransformTask,
            {"managed": False, "namespace_inject": "test", "api_version": "47.0"},
        )

        task._get_entities = mock.Mock()
        task._get_entities.return_value = {
            "CustomObject": {"Account", "Contact"},
            "ApexClass": {"Test"},
        }

        assert (
            task._generate_package_xml(False)
            == """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>Account</members>
        <members>Contact</members>
        <name>CustomObject</name>
    </types>
    <types>
        <members>Test</members>
        <name>ApexClass</name>
    </types>

    <version>47.0</version>
</Package>
"""
        )


class ConcreteMetadataSingleEntityTransformTask(MetadataSingleEntityTransformTask):
    def _transform_entity(self, xml_tree, api_name):
        return xml_tree


class TestMetadataSingleEntityTransformTask:
    def test_init_options(self):
        task = create_task(ConcreteMetadataSingleEntityTransformTask, {})
        task._init_options(
            {
                "managed": True,
                "api_version": "47.0",
                "namespace_inject": "test",
                "api_names": "%%%NAMESPACE%%%bar,foo",
            }
        )
        assert task.api_names == set(["test__bar", "foo"])

    def test_get_entities(self):
        task = create_task(
            ConcreteMetadataSingleEntityTransformTask,
            {"managed": False, "api_version": "47.0", "api_names": "bar,foo"},
        )

        assert task._get_entities() == {None: {"bar", "foo"}}

        task = create_task(
            ConcreteMetadataSingleEntityTransformTask,
            {"managed": False, "api_version": "47.0"},
        )

        assert task._get_entities() == {None: {"*"}}

    def test_transform(self):
        task = create_task(
            ConcreteMetadataSingleEntityTransformTask,
            {"managed": False, "api_version": "47.0", "api_names": "Test"},
        )

        task.entity = "CustomApplication"
        fake_tree = mock.Mock()
        fake_tree.tostring = mock.Mock(return_value="<pretend-xml/>")
        task._transform_entity = mock.Mock(return_value=fake_tree)

        input_xml = """<?xml version="1.0" encoding="UTF-8"?>
<CustomApplication xmlns="http://soap.sforce.com/2006/04/metadata">
</CustomApplication>"""

        with tempfile.TemporaryDirectory() as tmpdir:
            task._create_directories(tmpdir)

            test_path = task.retrieve_dir / "applications"
            test_path.mkdir()
            test_path = test_path / "Test.app"

            test_path.write_text(input_xml)

            task._transform()

            assert len(task._transform_entity.call_args_list) == 1

    def test_transform__bad_entity(self):
        task = create_task(
            ConcreteMetadataSingleEntityTransformTask,
            {"managed": False, "api_version": "47.0", "api_names": "bar,foo"},
        )

        task.entity = "Battlestar"

        with pytest.raises(CumulusCIException):
            task._transform()

    def test_transform__wildcard(self):
        task = create_task(
            ConcreteMetadataSingleEntityTransformTask,
            {"managed": False, "api_version": "47.0"},
        )

        task.entity = "CustomApplication"
        assert task.api_names == set("*")

        input_xml = """<?xml version="1.0" encoding="UTF-8"?>
<CustomApplication xmlns="http://soap.sforce.com/2006/04/metadata">
</CustomApplication>"""

        with tempfile.TemporaryDirectory() as tmpdir:
            task._create_directories(tmpdir)

            app_path = task.retrieve_dir / "applications"
            app_path.mkdir()
            test_path = app_path / "Test.app"
            test_path.write_text(input_xml)

            test_path = app_path / "Test_2.app"
            test_path.write_text(input_xml)

            task._transform()

            assert task.api_names == set(["Test", "Test_2"])
            assert (task.deploy_dir / "applications" / "Test.app").exists()
            assert (task.deploy_dir / "applications" / "Test_2.app").exists()

    def test_transform__remove_entity(self):
        task = create_task(
            ConcreteMetadataSingleEntityTransformTask,
            {"managed": False, "api_version": "47.0", "api_names": "*"},
        )

        task.entity = "CustomApplication"
        task._transform_entity = mock.Mock(
            side_effect=lambda xml, api_name: None if api_name == "Test_2" else xml
        )

        input_xml = """<?xml version="1.0" encoding="UTF-8"?>
<CustomApplication xmlns="http://soap.sforce.com/2006/04/metadata">
</CustomApplication>"""

        with tempfile.TemporaryDirectory() as tmpdir:
            task._create_directories(tmpdir)

            app_path = task.retrieve_dir / "applications"
            app_path.mkdir()
            test_path = app_path / "Test.app"
            test_path.write_text(input_xml)

            test_path = app_path / "Test_2.app"
            test_path.write_text(input_xml)

            task._transform()

            assert task.api_names == set(["Test"])
            assert (task.deploy_dir / "applications" / "Test.app").exists()
            assert not (task.deploy_dir / "applications" / "Test_2.app").exists()

    def test_transform__encoded_page_layout(self):
        task = create_task(
            ConcreteMetadataSingleEntityTransformTask,
            {"managed": False, "api_version": "47.0", "api_names": "*"},
        )

        task.entity = "Layout"
        task._transform_entity = mock.Mock(side_effect=lambda xml, api_name: xml)

        input_xml = """<?xml version="1.0" encoding="UTF-8"?>
<PageLayout xmlns="http://soap.sforce.com/2006/04/metadata">
</PageLayout>"""

        with tempfile.TemporaryDirectory() as tmpdir:
            task._create_directories(tmpdir)

            app_path = task.retrieve_dir / "layouts"
            app_path.mkdir()
            test_path = app_path / "Contact %28Marketing%29 Layout.layout"
            test_path.write_text(input_xml)

            task._transform()
            assert len(task._transform_entity.call_args_list) == 1
            assert (
                task._transform_entity.call_args_list[0][0][1]
                == "Contact (Marketing) Layout"
            )

            assert task.api_names == set(["Contact %28Marketing%29 Layout"])
            assert (
                task.deploy_dir / "layouts" / "Contact %28Marketing%29 Layout.layout"
            ).exists()

    def test_transform__non_xml_entity(self):
        task = create_task(
            ConcreteMetadataSingleEntityTransformTask,
            {"managed": False, "api_version": "47.0", "api_names": "bar,foo"},
        )

        task.entity = "LightningComponentBundle"

        with pytest.raises(CumulusCIException):
            task._transform()

    def test_transform__missing_record(self):
        task = create_task(
            ConcreteMetadataSingleEntityTransformTask,
            {"managed": False, "api_version": "47.0", "api_names": "Test"},
        )

        task.entity = "CustomApplication"

        with tempfile.TemporaryDirectory() as tmpdir:
            task._create_directories(tmpdir)

            test_path = task.retrieve_dir / "applications"
            test_path.mkdir()

            with pytest.raises(CumulusCIException):
                task._transform()

    def test_transform__xml_parse_error(self):
        task = create_task(
            ConcreteMetadataSingleEntityTransformTask,
            {"managed": False, "api_version": "47.0", "api_names": "Test"},
        )

        task.entity = "CustomApplication"

        with tempfile.TemporaryDirectory() as tmpdir:
            task._create_directories(tmpdir)

            test_path = task.retrieve_dir / "applications"
            test_path.mkdir()
            test_path = test_path / "Test.app"

            test_path.write_text(">>>>>NOT XML<<<<<")
            with pytest.raises(etree.ParseError):
                task._transform()


class TestUpdateMetadataFirstChildTextTask:
    def test_init_options__namespace_injected_in_value(self):
        options = {
            "managed": True,
            "namespace_inject": "namespace",
            "metadata_type": "CustomObject",
            "tag": "customObjectAttribute",
            "value": "%%%NAMESPACE%%%newAttributeValue",
        }

        expected_value = "namespace__newAttributeValue"

        task = create_task(UpdateMetadataFirstChildTextTask, options)

        assert expected_value == task.options["value"]

        # task.entity should always be set as options["metadata_type"]
        assert options["metadata_type"] == task.entity

    def test_transform_entity__child_found(self):
        api_name = "Supercalifragilisticexpialidocious__c"

        entity = "CustomObject"
        tag = "customObjectAttribute"
        value = "newAttributeValue"

        ORIGINAL_XML = f"""<?xml version="1.0" encoding="UTF-8"?>
<{entity} xmlns="http://soap.sforce.com/2006/04/metadata">
    <name>{api_name}</name>
    <{tag}>oldAttributeValue</{tag}>
    <anotherTag>value</anotherTag>
</{entity}>
"""

        EXPECTED_XML = f"""<?xml version="1.0" encoding="UTF-8"?>
<{entity} xmlns="http://soap.sforce.com/2006/04/metadata">
    <name>{api_name}</name>
    <{tag}>{value}</{tag}>
    <anotherTag>value</anotherTag>
</{entity}>
"""

        metadata = fromstring(ORIGINAL_XML.encode("utf-8"))

        task = create_task(
            UpdateMetadataFirstChildTextTask,
            {
                "managed": False,
                "namespace_inject": None,
                "metadata_type": entity,
                "tag": tag,
                "value": value,
            },
        )
        task.logger = mock.Mock()
        assert tag == task.options.get("tag")
        assert value == task.options.get("value")

        actual = task._transform_entity(metadata, api_name)

        assert metadata == actual

        assert actual.tostring(xml_declaration=True) == EXPECTED_XML

        task.logger.info.assert_has_calls(
            [
                mock.call(f'Updating {entity} "{api_name}":'),
                mock.call(f'    {tag} as "{value}"'),
            ]
        )

    def test_transform_entity__child_not_found(self):
        api_name = "Supercalifragilisticexpialidocious__c"

        entity = "CustomObject"
        tag = "customObjectAttribute"
        value = "newAttributeValue"

        ORIGINAL_XML = f"""<?xml version="1.0" encoding="UTF-8"?>
<{entity} xmlns="http://soap.sforce.com/2006/04/metadata">
    <name>{api_name}</name>
    <anotherTag>value</anotherTag>
</{entity}>
"""

        EXPECTED_XML = f"""<?xml version="1.0" encoding="UTF-8"?>
<{entity} xmlns="http://soap.sforce.com/2006/04/metadata">
    <name>{api_name}</name>
    <anotherTag>value</anotherTag>
    <{tag}>{value}</{tag}>
</{entity}>
"""

        metadata = fromstring(ORIGINAL_XML.encode("utf-8"))

        expected_metadata = fromstring(EXPECTED_XML.encode("utf-8"))

        task = create_task(
            UpdateMetadataFirstChildTextTask,
            {
                "managed": False,
                "namespace_inject": None,
                "metadata_type": entity,
                "tag": tag,
                "value": value,
            },
        )
        task.logger = mock.Mock()
        assert tag == task.options.get("tag")
        assert value == task.options.get("value")

        actual = task._transform_entity(metadata, api_name)

        assert metadata == actual

        assert actual.tostring() == expected_metadata.tostring()

        task.logger.info.assert_has_calls(
            [
                mock.call(f'Updating {entity} "{api_name}":'),
                mock.call(f'    {tag} as "{value}"'),
            ]
        )
